from typing import Any, Callable, List, Optional, Tuple, Union

import numpy as np
import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.math import greater_equal
from tensorflow.types.experimental import TensorLike

from decomon.core import (
    BoxDomain,
    ForwardMode,
    GridDomain,
    InputsOutputsSpec,
    PerturbationDomain,
    Slope,
    get_affine,
    get_ibp,
)

TensorFunction = Callable[[TensorLike], tf.Tensor]


# linear hull for activation function
def relu_prime(x: TensorLike) -> tf.Tensor:
    """Derivative of relu

    Args:
        x

    Returns:

    """

    return K.clip(K.sign(x), K.cast(0, dtype=x.dtype), K.cast(1, dtype=x.dtype))


def sigmoid_prime(x: TensorLike) -> tf.Tensor:
    """Derivative of sigmoid

    Args:
        x

    Returns:

    """

    s_x = K.sigmoid(x)
    return s_x * (K.cast(1, dtype=x.dtype) - s_x)


def tanh_prime(x: TensorLike) -> tf.Tensor:
    """Derivative of tanh

    Args:
        x

    Returns:

    """

    s_x = K.tanh(x)
    return K.cast(1, dtype=x.dtype) - K.pow(s_x, K.cast(2, dtype=x.dtype))


def softsign_prime(x: TensorLike) -> tf.Tensor:
    """Derivative of softsign

    Args:
        x

    Returns:

    """

    return K.cast(1.0, dtype=x.dtype) / K.pow(K.cast(1.0, dtype=x.dtype) + K.abs(x), K.cast(2, dtype=x.dtype))


##### corners ######
def get_lower_bound_grid(x: tf.Tensor, W: tf.Tensor, b: tf.Tensor, n: int) -> tf.Tensor:
    A, B = convert_lower_search_2_subset_sum(x, W, b, n)
    return subset_sum_lower(A, B, repeat=n)


def get_upper_bound_grid(x: tf.Tensor, W: tf.Tensor, b: tf.Tensor, n: int) -> tf.Tensor:
    return -get_lower_bound_grid(x, -W, -b, n)


def get_bound_grid(
    x: tf.Tensor, W_u: tf.Tensor, b_u: tf.Tensor, W_l: tf.Tensor, b_l: tf.Tensor, n: int
) -> Tuple[tf.Tensor, tf.Tensor]:
    upper = get_upper_bound_grid(x, W_u, b_u, n)
    lower = get_lower_bound_grid(x, W_l, b_l, n)

    return upper, lower


# convert max Wx +b s.t Wx+b<=0 into a subset-sum problem with positive values
def convert_lower_search_2_subset_sum(x: tf.Tensor, W: tf.Tensor, b: tf.Tensor, n: int) -> Tuple[tf.Tensor, tf.Tensor]:
    x_min = x[:, 0]
    x_max = x[:, 1]

    if len(W.shape) > 3:
        W = K.reshape(W, (-1, W.shape[1], np.prod(W.shape[2:])))
        b = K.reshape(b, (-1, np.prod(b.shape[1:])))

    const = BoxDomain().get_lower(x, W, b)

    weights = K.abs(W) * K.expand_dims((x_max - x_min) / n, -1)
    return weights, const


def subset_sum_lower(W: tf.Tensor, b: tf.Tensor, repeat: int = 1) -> tf.Tensor:
    B = tf.sort(W, 1)
    C = K.repeat_elements(B, rep=repeat, axis=1)
    C_reduced = K.cumsum(C, axis=1)
    D = K.minimum(K.sign(K.expand_dims(-b, 1) - C_reduced) + 1, 1)

    score = K.minimum(K.sum(D * C, 1) + b, 0.0)
    return score


# define routines to get linear relaxations useful both for forward and backward
def get_linear_hull_relu(
    upper: tf.Tensor,
    lower: tf.Tensor,
    slope: Union[str, Slope],
    upper_g: float = 0.0,
    lower_g: float = 0.0,
    **kwargs: Any,
) -> List[tf.Tensor]:
    slope = Slope(slope)
    # in case upper=lower, this cases are
    # considered with index_dead and index_linear
    alpha = (K.relu(upper) - K.relu(lower)) / K.maximum(K.cast(K.epsilon(), dtype=upper.dtype), upper - lower)

    # scaling factor for the upper bound on the relu
    # see README

    w_u = alpha
    b_u = K.relu(lower) - alpha * lower
    z_value = K.cast(0.0, dtype=upper.dtype)
    o_value = K.cast(1.0, dtype=upper.dtype)

    if slope == Slope.V_SLOPE:
        # 1 if upper<=-lower else 0
        index_a = -K.clip(K.sign(upper + lower) - o_value, -o_value, z_value)

        # 1 if upper>-lower else 0
        index_b = o_value - index_a
        w_l = index_b
        b_l = z_value * b_u

    elif slope == Slope.A_SLOPE:
        w_l = K.clip(K.sign(w_u - 0.5), 0, 1)
        b_l = z_value * b_u

    elif slope == Slope.Z_SLOPE:
        w_l = z_value * w_u
        b_l = z_value * b_u

    elif slope == Slope.O_SLOPE:
        w_l = z_value * w_u + o_value
        b_l = z_value * b_u

    elif slope == Slope.S_SLOPE:
        w_l = w_u
        b_l = z_value * b_u

    else:
        raise NotImplementedError(f"Not implemented for slope {slope}")

    if "upper_grid" in kwargs:
        raise NotImplementedError()

    gamma = o_value
    if "finetune" in kwargs:
        # retrieve variables to optimize the slopes
        gamma = kwargs["finetune"][None]

    w_l = gamma * w_l + (o_value - gamma) * (o_value - w_l)

    # check inactive relu state: u<=0
    index_dead = -K.clip(K.sign(upper) - o_value, -o_value, z_value)  # =1 if inactive state
    index_linear = K.clip(K.sign(lower) + o_value, z_value, o_value)  # 1 if linear state

    w_u = (o_value - index_dead) * w_u
    w_l = (o_value - index_dead) * w_l
    b_u = (o_value - index_dead) * b_u
    b_l = (o_value - index_dead) * b_l

    w_u = (o_value - index_linear) * w_u + index_linear
    w_l = (o_value - index_linear) * w_l + index_linear
    b_u = (o_value - index_linear) * b_u
    b_l = (o_value - index_linear) * b_l

    return [w_u, b_u, w_l, b_l]


def get_linear_hull_sigmoid(
    upper: tf.Tensor, lower: tf.Tensor, slope: Union[str, Slope], **kwargs: Any
) -> List[tf.Tensor]:
    x = [upper, lower]
    return get_linear_hull_s_shape(x, func=K.sigmoid, f_prime=sigmoid_prime, mode=ForwardMode.IBP, **kwargs)


def get_linear_hull_tanh(
    upper: tf.Tensor, lower: tf.Tensor, slope: Union[str, Slope], **kwargs: Any
) -> List[tf.Tensor]:
    x = [upper, lower]
    return get_linear_hull_s_shape(x, func=K.tanh, f_prime=tanh_prime, mode=ForwardMode.IBP, **kwargs)


def get_linear_softplus_hull(
    upper: tf.Tensor, lower: tf.Tensor, slope: Union[str, Slope], **kwargs: Any
) -> List[tf.Tensor]:
    slope = Slope(slope)
    # in case upper=lower, this cases are
    # considered with index_dead and index_linear
    u_c = K.softsign(upper)
    l_c = K.softsign(lower)
    alpha = (u_c - l_c) / K.maximum(K.cast(K.epsilon(), dtype=upper.dtype), (upper - lower))
    w_u = alpha
    b_u = -alpha * lower + l_c

    z_value = K.cast(0.0, dtype=upper.dtype)
    o_value = K.cast(1.0, dtype=upper.dtype)

    if slope == Slope.V_SLOPE:
        # 1 if upper<=-lower else 0
        index_a = -K.clip(K.sign(upper + lower) - o_value, -o_value, z_value)
        # 1 if upper>-lower else 0
        index_b = o_value - index_a
        w_l = index_b
        b_l = z_value * b_u
    elif slope == Slope.Z_SLOPE:
        w_l = z_value * w_u
        b_l = z_value * b_u
    elif slope == Slope.O_SLOPE:
        w_l = z_value * w_u + o_value
        b_l = z_value * b_u
    elif slope == Slope.S_SLOPE:
        w_l = w_u
        b_l = z_value * b_u
    else:
        raise ValueError(f"Unknown slope {slope}")

    index_dead = -K.clip(K.sign(upper) - o_value, -o_value, z_value)

    w_u = (o_value - index_dead) * w_u
    w_l = (o_value - index_dead) * w_l
    b_u = (o_value - index_dead) * b_u
    b_l = (o_value - index_dead) * b_l

    if "finetune" in kwargs:
        # weighted linear combination
        alpha_u, alpha_l = kwargs["finetune"]
        alpha_u = alpha_u[None]
        alpha_l = alpha_l[None]

        w_u = alpha_u * w_u
        b_u = alpha_u * b_u + (o_value - alpha_u) * K.maximum(upper, z_value)

        w_l = alpha_l * w_l
        b_l = alpha_l * b_l + (o_value - alpha_l) * K.maximum(lower, z_value)

    return [w_u, b_u, w_l, b_l]


def subtract(
    inputs_0: List[tf.Tensor],
    inputs_1: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
) -> List[tf.Tensor]:
    """LiRPA implementation of inputs_0-inputs_1

    Args:
        inputs_0: tensor
        inputs_1: tensor
        dc_decomp: boolean that indicates
        perturbation_domain: the type of perturbation domain
    whether we return a difference of convex decomposition of our layer

    Returns:
        inputs_0 - inputs_1
    """
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    inputs_1 = minus(inputs_1, mode=mode, dc_decomp=dc_decomp, perturbation_domain=perturbation_domain)
    output = add(inputs_0, inputs_1, dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)

    return output


def add(
    inputs_0: List[tf.Tensor],
    inputs_1: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
) -> List[tf.Tensor]:
    """LiRPA implementation of inputs_0+inputs_1

    Args:
        inputs_0: tensor
        inputs_1: tensor
        dc_decomp: boolean that indicates
        perturbation_domain: the type of perturbation domain
    whether we return a difference of convex decomposition of our layer

    Returns:
        inputs_0 + inputs_1
    """
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()

    mode = ForwardMode(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)

    x_0, u_c_0, w_u_0, b_u_0, l_c_0, w_l_0, b_l_0, h_0, g_0 = inputs_outputs_spec.get_fullinputs_from_inputsformode(
        inputs_0
    )
    _, u_c_1, w_u_1, b_u_1, l_c_1, w_l_1, b_l_1, h_1, g_1 = inputs_outputs_spec.get_fullinputs_from_inputsformode(
        inputs_1
    )

    x = x_0
    h = h_0 + h_1
    g = g_0 + g_1
    u_c = u_c_0 + u_c_1
    l_c = l_c_0 + l_c_1
    w_u = w_u_0 + w_u_1
    w_l = w_l_0 + w_l_1
    b_u = b_u_0 + b_u_1
    b_l = b_l_0 + b_l_1

    fulloutputs = [x, u_c, w_u, b_u, l_c, w_l, b_l, h, g]
    outputs = inputs_outputs_spec.extract_outputsformode_from_fulloutputs(fulloutputs)

    return outputs


def relu_(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    **kwargs: Any,
) -> List[tf.Tensor]:
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()

    mode = ForwardMode(mode)
    affine = get_affine(mode)
    ibp = get_ibp(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)

    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(inputs)
    dtype = x.dtype

    z_value = K.cast(0.0, dtype=dtype)
    o_value = K.cast(1.0, dtype=dtype)
    empty_tensor = inputs_outputs_spec.get_empty_tensor(dtype=dtype)

    if dc_decomp:
        h_out = K.maximum(h, -g)
        g_out = g
        index_dead = -K.clip(K.sign(u_c) - o_value, -o_value, z_value)  # =1 if inactive state
        index_linear = K.clip(K.sign(l_c) + o_value, z_value, o_value)  # 1 if linear state

        h_out = (o_value - index_dead) * h_out
        g_out = (o_value - index_dead) * g_out
        h_out = (o_value - index_linear) * h_out + index_linear * h
        g_out = (o_value - index_linear) * g_out + index_linear * g
    else:
        h_out, g_out = empty_tensor, empty_tensor

    if ibp:
        u_c_out = K.relu(u_c)
        l_c_out = K.relu(l_c)
    else:
        u_c_out = empty_tensor
        l_c_out = empty_tensor

    if affine:
        if isinstance(perturbation_domain, GridDomain):
            upper_g, lower_g = get_bound_grid(x, w_u, b_u, w_l, b_l, 1)
            kwargs.update({"upper_grid": upper_g, "lower_grid": lower_g})

        w_u_out, b_u_out, w_l_out, b_l_out = get_linear_hull_relu(u_c, l_c, slope, **kwargs)
        b_u_out = w_u_out * b_u + b_u_out
        b_l_out = w_l_out * b_l + b_l_out
        w_u_out = K.expand_dims(w_u_out, 1) * w_u
        w_l_out = K.expand_dims(w_l_out, 1) * w_l
    else:
        w_u_out, b_u_out, w_l_out, b_l_out = empty_tensor, empty_tensor, empty_tensor, empty_tensor

    return inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
        [x, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out, h_out, g_out]
    )


def minus(
    inputs: List[tf.Tensor],
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """LiRPA implementation of minus(x)=-x.

    Args:
        inputs
        mode

    Returns:

    """
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    mode = ForwardMode(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)

    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(
        inputs, tight=False, compute_ibp_from_affine=False
    )

    u_c_out = -l_c
    l_c_out = -u_c
    w_u_out = -w_l
    b_u_out = -b_l
    w_l_out = -w_u
    b_l_out = -b_u
    h_out = -g
    g_out = -h

    return inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
        [x, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out, h_out, g_out]
    )


def maximum(
    inputs_0: List[tf.Tensor],
    inputs_1: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    finetune: bool = False,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """LiRPA implementation of element-wise max

    Args:
        inputs_0: list of tensors
        inputs_1: list of tensors
        dc_decomp: boolean that indicates
        perturbation_domain: the type of perturbation domain
    whether we return a difference of convex decomposition of our layer

    Returns:
        maximum(inputs_0, inputs_1)
    """
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    output_0 = subtract(inputs_1, inputs_0, dc_decomp=dc_decomp, perturbation_domain=perturbation_domain, mode=mode)
    if finetune:
        finetune = kwargs["finetune_params"]
        output_1 = relu_(
            output_0, dc_decomp=dc_decomp, perturbation_domain=perturbation_domain, mode=mode, finetune=finetune
        )
    else:
        output_1 = relu_(output_0, dc_decomp=dc_decomp, perturbation_domain=perturbation_domain, mode=mode)

    return add(
        output_1,
        inputs_0,
        dc_decomp=dc_decomp,
        perturbation_domain=perturbation_domain,
        mode=mode,
    )


def minimum(
    inputs_0: List[tf.Tensor],
    inputs_1: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    finetune: bool = False,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """LiRPA implementation of element-wise min

    Args:
        inputs_0
        inputs_1
        dc_decomp
        perturbation_domain
        mode

    Returns:

    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    return minus(
        maximum(
            minus(inputs_0, dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain),
            minus(inputs_1, dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain),
            dc_decomp=dc_decomp,
            perturbation_domain=perturbation_domain,
            mode=mode,
            finetune=finetune,
            **kwargs,
        ),
        dc_decomp=dc_decomp,
        mode=mode,
        perturbation_domain=perturbation_domain,
    )


def get_linear_hull_s_shape(
    inputs: List[tf.Tensor],
    func: TensorFunction = K.sigmoid,
    f_prime: TensorFunction = sigmoid_prime,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    dc_decomp: bool = False,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """Computing the linear hull of shape functions  given the pre activation neurons

    Args:
        inputs: list of input tensors
        func: the function (sigmoid, tanh, softsign...)
        f_prime: the derivative of the function (sigmoid_prime...)
        perturbation_domain: the type of convex input domain
        mode: type of Forward propagation (ibp, affine, or hybrid)

    Returns:
        the updated list of tensors
    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    mode = ForwardMode(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)

    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(inputs)
    dtype = x.dtype

    z_value = K.cast(0.0, dtype=dtype)
    o_value = K.cast(1.0, dtype=dtype)
    t_value = K.cast(2.0, dtype=dtype)

    # flatten
    shape = list(u_c.shape[1:])
    u_c_flat = K.reshape(u_c, (-1, np.prod(shape)))  # (None, n)
    l_c_flat = K.reshape(l_c, (-1, np.prod(shape)))  # (None, n)

    # upper bound
    # derivative
    s_u_prime = f_prime(u_c_flat)  # (None, n)
    s_l_prime = f_prime(l_c_flat)  # (None, n)
    s_u = func(u_c_flat)  # (None, n)
    s_l = func(l_c_flat)  # (None, n)

    # case 0:
    coeff = (s_u - s_l) / K.maximum(K.cast(K.epsilon(), dtype=dtype), u_c_flat - l_c_flat)
    alpha_u_0 = K.switch(greater_equal(s_u_prime, coeff), o_value + z_value * u_c_flat, z_value * u_c_flat)  # (None, n)
    alpha_u_1 = (o_value - alpha_u_0) * ((K.sign(l_c_flat) + o_value) / t_value)

    w_u_0 = coeff
    b_u_0 = -w_u_0 * l_c_flat + s_l

    w_u_1 = z_value * u_c_flat
    b_u_1 = s_u

    w_u_2, b_u_2 = get_t_upper(u_c_flat, l_c_flat, s_l, func=func, f_prime=f_prime)

    w_u_out = K.reshape(alpha_u_0 * w_u_0 + alpha_u_1 * w_u_1 + (o_value - alpha_u_0 - alpha_u_1) * w_u_2, [-1] + shape)
    b_u_out = K.reshape(alpha_u_0 * b_u_0 + alpha_u_1 * b_u_1 + (o_value - alpha_u_0 - alpha_u_1) * b_u_2, [-1] + shape)

    # linear hull
    # case 0:
    alpha_l_0 = K.switch(greater_equal(s_l_prime, coeff), o_value + z_value * l_c_flat, z_value * l_c_flat)  # (None, n)
    alpha_l_1 = (o_value - alpha_l_0) * ((K.sign(-u_c_flat) + o_value) / t_value)

    w_l_0 = coeff
    b_l_0 = -w_l_0 * u_c_flat + s_u

    w_l_1 = z_value * u_c_flat
    b_l_1 = s_l

    w_l_2, b_l_2 = get_t_lower(u_c_flat, l_c_flat, s_u, func=func, f_prime=f_prime)

    w_l_out = K.reshape(alpha_l_0 * w_l_0 + alpha_l_1 * w_l_1 + (o_value - alpha_l_0 - alpha_l_1) * w_l_2, [-1] + shape)
    b_l_out = K.reshape(alpha_l_0 * b_l_0 + alpha_l_1 * b_l_1 + (o_value - alpha_l_0 - alpha_l_1) * b_l_2, [-1] + shape)

    return [w_u_out, b_u_out, w_l_out, b_l_out]


def get_t_upper(
    u_c_flat: tf.Tensor,
    l_c_flat: tf.Tensor,
    s_l: tf.Tensor,
    func: TensorFunction = K.sigmoid,
    f_prime: TensorFunction = sigmoid_prime,
) -> List[tf.Tensor]:
    """linear interpolation between lower and upper bounds on the function func to have a symbolic approximation of the best
    coefficient for the affine upper bound

    Args:
        u_c_flat: flatten tensor of constant upper bound
        l_c_flat: flatten tensor of constant lower bound
        s_l: lowest value of the function func on the domain
        func: the function (sigmoid, tanh,  softsign)
        f_prime: the derivative of the function

    Returns:
        the upper affine bounds in this subcase
    """

    o_value = K.cast(1.0, dtype=u_c_flat.dtype)
    z_value = K.cast(0.0, dtype=u_c_flat.dtype)

    # step1: find t
    u_c_reshaped = K.expand_dims(u_c_flat, -1)  # (None, n , 1)
    l_c_reshaped = K.expand_dims(l_c_flat, -1)  # (None, n,  1)
    t = K.cast(np.linspace(0, 1, 100)[None, None, :], dtype=u_c_flat.dtype) * u_c_reshaped  # (None, n , 100)

    s_p_t = f_prime(t)  # (None, n, 100)
    s_t = func(t)  # (None, n, 100)

    score = K.abs(s_p_t - (s_t - K.expand_dims(s_l, -1)) / (t - l_c_reshaped))  # (None, n, 100)
    index = K.argmin(score, -1)  # (None, n)
    threshold = K.min(score, -1)  # (None, n)

    index_t = K.cast(
        K.switch(K.greater(threshold, z_value * threshold), index, K.clip(index - 1, 0, 100)), dtype=u_c_flat.dtype
    )  # (None, n)
    t_value = K.sum(
        K.switch(
            K.equal(
                o_value * K.cast(np.arange(0, 100)[None, None, :], dtype=u_c_flat.dtype) + z_value * u_c_reshaped,
                K.expand_dims(index_t, -1) + z_value * u_c_reshaped,
            ),
            t,
            z_value * t,
        ),
        -1,
    )  # (None, n)

    s_t = func(t_value)  # (None, n)
    w_u = (s_t - s_l) / K.maximum(K.cast(K.epsilon(), dtype=u_c_flat.dtype), t_value - l_c_flat)  # (None, n)
    b_u = -w_u * l_c_flat + s_l  # + func(l_c_flat)

    return [w_u, b_u]


def get_t_lower(
    u_c_flat: tf.Tensor,
    l_c_flat: tf.Tensor,
    s_u: tf.Tensor,
    func: TensorFunction = K.sigmoid,
    f_prime: TensorFunction = sigmoid_prime,
) -> List[tf.Tensor]:
    """linear interpolation between lower and upper bounds on the function func to have a symbolic approximation of the best
    coefficient for the affine lower bound

    Args:
        u_c_flat: flatten tensor of constant upper bound
        l_c_flat: flatten tensor of constant lower bound
        s_u: highest value of the function func on the domain
        func: the function (sigmoid, tanh,  softsign)
        f_prime: the derivative of the function

    Returns:
        the lower affine bounds in this subcase
    """
    z_value = K.cast(0.0, dtype=u_c_flat.dtype)
    o_value = K.cast(1.0, dtype=u_c_flat.dtype)

    # step1: find t
    u_c_reshaped = K.expand_dims(u_c_flat, -1)  # (None, n , 1)
    l_c_reshaped = K.expand_dims(l_c_flat, -1)  # (None, n,  1)
    t = K.cast(np.linspace(0, 1.0, 100)[None, None, :], dtype=u_c_flat.dtype) * l_c_reshaped  # (None, n , 100)

    s_p_t = f_prime(t)  # (None, n, 100)
    s_t = func(t)  # (None, n, 100)

    score = K.abs(s_p_t - (K.expand_dims(s_u, -1) - s_t) / (u_c_reshaped - t))  # (None, n, 100)
    index = K.argmin(score, -1)  # (None, n)

    threshold = K.min(score, -1)
    index_t = K.cast(
        K.switch(K.greater(threshold, z_value * threshold), index, K.clip(index + 1, 0, 100)), dtype=u_c_flat.dtype
    )  # (None, n)
    t_value = K.sum(
        K.switch(
            K.equal(
                o_value * K.cast(np.arange(0, 100)[None, None, :], dtype=u_c_flat.dtype) + z_value * u_c_reshaped,
                K.expand_dims(index_t, -1) + z_value * u_c_reshaped,
            ),
            t,
            z_value * t,
        ),
        -1,
    )

    s_t = func(t_value)  # (None, n)
    w_l = (s_u - s_t) / K.maximum(K.cast(K.epsilon(), dtype=u_c_flat.dtype), u_c_flat - t_value)  # (None, n)
    b_l = -w_l * u_c_flat + s_u  # func(u_c_flat)

    return [w_l, b_l]



import pytest
from tensorflow.keras.layers import Layer

from decomon.backward_layers.backward_layers import (
    BackwardActivation,
    BackwardBatchNormalization,
    BackwardConv2D,
    BackwardDense,
    BackwardDropout,
    BackwardFlatten,
    BackwardInputLayer,
    BackwardPermute,
    BackwardReshape,
)
from decomon.backward_layers.backward_maxpooling import BackwardMaxPooling2D
from decomon.backward_layers.backward_merge import (
    BackwardAdd,
    BackwardAverage,
    BackwardConcatenate,
    BackwardMaximum,
    BackwardMinimum,
    BackwardMultiply,
    BackwardSubtract,
)
from decomon.backward_layers.deel_lip import BackwardGroupSort2
from decomon.layers.decomon_layers import (
    DecomonActivation,
    DecomonBatchNormalization,
    DecomonConv2D,
    DecomonDense,
    DecomonDropout,
)
from decomon.layers.decomon_merge_layers import (
    DecomonAdd,
    DecomonAverage,
    DecomonConcatenate,
    DecomonMaximum,
    DecomonMinimum,
    DecomonMultiply,
    DecomonSubtract,
)
from decomon.layers.decomon_reshape import DecomonPermute, DecomonReshape
from decomon.layers.maxpooling import DecomonMaxPooling2D

try:
    from decomon.layers.deel_lip import DecomonGroupSort2
except ImportError:
    deel_lip_available = False
else:
    deel_lip_available = True


deel_lip_skip_reason = "deel-lip is not available"


def test_backward_layers():
    activation = "linear"
    sublayer = DecomonActivation(activation)
    layer = BackwardActivation(layer=sublayer)
    config = layer.get_config()
    assert config["layer"]["class_name"] == sublayer.__class__.__name__

    units = 2
    sublayer = DecomonDense(units=units, use_bias=False)
    layer = BackwardDense(layer=sublayer)
    config = layer.get_config()
    assert config["layer"]["class_name"] == sublayer.__class__.__name__

    filters = 2
    kernel_size = 3, 3
    sublayer = DecomonConv2D(filters=filters, kernel_size=kernel_size)
    layer = BackwardConv2D(layer=sublayer)
    config = layer.get_config()
    assert config["layer"]["class_name"] == sublayer.__class__.__name__

    rate = 0.9
    sublayer = DecomonDropout(rate=rate)
    layer = BackwardDropout(layer=sublayer)
    config = layer.get_config()
    assert config["layer"]["class_name"] == sublayer.__class__.__name__

    dims = (1, 2, 3)
    sublayer = DecomonPermute(dims=dims)
    layer = BackwardPermute(layer=sublayer)
    config = layer.get_config()
    assert config["layer"]["class_name"] == sublayer.__class__.__name__

    shape = (1, 2, 3)
    sublayer = DecomonReshape(target_shape=shape)
    layer = BackwardReshape(layer=sublayer)
    config = layer.get_config()
    assert config["layer"]["class_name"] == sublayer.__class__.__name__

    sublayer = DecomonBatchNormalization()
    layer = BackwardBatchNormalization(layer=sublayer)
    config = layer.get_config()
    assert config["layer"]["class_name"] == sublayer.__class__.__name__

    sublayer = Layer()
    for cls in [BackwardFlatten, BackwardInputLayer]:
        layer = cls(layer=sublayer)
        config = layer.get_config()
        assert config["layer"]["class_name"] == sublayer.__class__.__name__


backward2decomon = {
    BackwardAdd: DecomonAdd,
    BackwardAverage: DecomonAverage,
    BackwardConcatenate: DecomonConcatenate,
    # BackwardDot: DecomonDot,  # not yet implemented
    BackwardMaximum: DecomonMaximum,
    BackwardMinimum: DecomonMinimum,
    BackwardMultiply: DecomonMultiply,
    BackwardSubtract: DecomonSubtract,
}


def test_backward_merge():
    for backwardlayer_cls, decomonlayer_cls in backward2decomon.items():
        sublayer = decomonlayer_cls()
        layer = backwardlayer_cls(layer=sublayer)
        config = layer.get_config()
        assert config["layer"]["class_name"] == sublayer.__class__.__name__


@pytest.mark.skip("Not yet implemented.")
def test_backward_maxpooling():
    sublayer = DecomonMaxPooling2D()
    layer = BackwardMaxPooling2D(layer=sublayer)
    config = layer.get_config()
    assert config["layer"]["class_name"] == sublayer.__class__.__name__


@pytest.mark.skip("Not yet implemented.")
@pytest.mark.skipif(not (deel_lip_available), reason=deel_lip_skip_reason)
def test_deel_lip():
    sublayer = DecomonGroupSort2()
    layer = BackwardGroupSort2(layer=sublayer)
    config = layer.get_config()
    assert config["layer"]["class_name"] == sublayer.__class__.__name__



# Test unit for decomon with Dense layers


import numpy as np
import pytest
import tensorflow.keras.backend as K
from tensorflow.python.keras.backend import _get_available_gpus

from decomon.backward_layers.convert import to_backward
from decomon.layers.decomon_layers import DecomonConv2D


def test_Decomon_conv_box(data_format, padding, use_bias, mode, floatx, decimal, helpers):
    if data_format == "channels_first" and not len(_get_available_gpus()):
        pytest.skip("data format 'channels first' is possible only in GPU mode")

    odd, m_0, m_1 = 0, 0, 1
    dc_decomp = False

    # tensor inputs
    inputs = helpers.get_tensor_decomposition_images_box(data_format, odd, dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs=inputs)

    # numpy inputs
    inputs_ = helpers.get_standard_values_images_box(data_format, odd, m0=m_0, m1=m_1, dc_decomp=dc_decomp)
    x, y, z, u_c, W_u, b_u, l_c, W_l, b_l = inputs_

    # decomon layer
    decomon_layer = DecomonConv2D(
        10,
        kernel_size=(3, 3),
        dc_decomp=dc_decomp,
        padding=padding,
        use_bias=use_bias,
        mode=mode,
        dtype=K.floatx(),
    )
    decomon_layer(inputs_for_mode)  # init weights

    # get backward layer
    backward_layer = to_backward(decomon_layer)

    # backward outputs
    outputs = backward_layer(inputs_for_mode)
    f_decomon = K.function(inputs, outputs)
    outputs_ = f_decomon(inputs_)

    w_u_, b_u_, w_l_, b_l_ = outputs_

    # reshape the matrices
    w_u_ = w_u_[:, None]
    w_l_ = w_l_[:, None]
    b_u_ = b_u_[:, None]
    b_l_ = b_l_[:, None]

    b_l = b_l.reshape((len(b_l), -1))
    b_u = b_u.reshape((len(b_u), -1))
    W_u = W_u.reshape((len(W_u), W_u.shape[1], -1))
    W_l = W_l.reshape((len(W_l), W_l.shape[1], -1))

    # backward recomposition
    w_r_u = np.sum(np.maximum(0.0, w_u_) * np.expand_dims(W_u, -1), 2) + np.sum(
        np.minimum(0.0, w_u_) * np.expand_dims(W_l, -1), 2
    )
    w_r_l = np.sum(np.maximum(0.0, w_l_) * np.expand_dims(W_l, -1), 2) + np.sum(
        np.minimum(0.0, w_l_) * np.expand_dims(W_u, -1), 2
    )
    b_r_u = (
        np.sum(np.maximum(0, w_u_[:, 0]) * np.expand_dims(b_u, -1), 1)[:, None]
        + np.sum(np.minimum(0, w_u_[:, 0]) * np.expand_dims(b_l, -1), 1)[:, None]
        + b_u_
    )
    b_r_l = (
        np.sum(np.maximum(0, w_l_[:, 0]) * np.expand_dims(b_l, -1), 1)[:, None]
        + np.sum(np.minimum(0, w_l_[:, 0]) * np.expand_dims(b_u, -1), 1)[:, None]
        + b_l_
    )

    # check bounds consistency
    helpers.assert_output_properties_box_linear(x, None, z[:, 0], z[:, 1], None, w_r_u, b_r_u, None, w_r_l, b_r_l)



from typing import Any, Dict, List, Optional, Tuple, Type, Union

import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.layers import InputSpec, Layer, Permute, Reshape

from decomon.core import ForwardMode, PerturbationDomain, get_affine, get_ibp
from decomon.layers.core import DecomonLayer


class DecomonReshape(DecomonLayer, Reshape):
    """Forward LiRPA implementation of Reshape layers.
    See Keras official documentation for further details on the Reshape operator
    """

    original_keras_layer_class = Reshape

    def __init__(
        self,
        target_shape: Tuple[int, ...],
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        finetune: bool = False,
        shared: bool = False,
        fast: bool = True,
        **kwargs: Any,
    ):
        """
        Args:
            data_format
            **kwargs
        """
        super().__init__(
            target_shape=target_shape,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            mode=mode,
            finetune=finetune,
            shared=shared,
            fast=fast,
            **kwargs,
        )

        if self.mode == ForwardMode.HYBRID:
            self.input_spec = [
                InputSpec(min_ndim=1),  # z
                InputSpec(min_ndim=1),  # u_c
                InputSpec(min_ndim=1),  # w_u
                InputSpec(min_ndim=1),  # b_u
                InputSpec(min_ndim=1),  # l_c
                InputSpec(min_ndim=1),  # w_l
                InputSpec(min_ndim=1),  # b_l
            ]
        elif self.mode == ForwardMode.IBP:
            self.input_spec = [
                InputSpec(min_ndim=1),  # u_c
                InputSpec(min_ndim=1),  # l_c
            ]
        if self.mode == ForwardMode.AFFINE:
            self.input_spec = [
                InputSpec(min_ndim=1),  # z
                InputSpec(min_ndim=1),  # w_u
                InputSpec(min_ndim=1),  # b_u
                InputSpec(min_ndim=1),  # w_l
                InputSpec(min_ndim=1),  # b_l
            ]

        if self.dc_decomp:
            self.input_spec += [InputSpec(min_ndim=1), InputSpec(min_ndim=1)]

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        def op(x: tf.Tensor) -> tf.Tensor:
            return Reshape.call(self, x)

        x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = self.inputs_outputs_spec.get_fullinputs_from_inputsformode(
            inputs, compute_ibp_from_affine=False
        )
        dtype = x.dtype
        empty_tensor = self.inputs_outputs_spec.get_empty_tensor(dtype=dtype)

        if self.dc_decomp:
            h_out = op(h)
            g_out = op(g)
        else:
            h_out, g_out = empty_tensor, empty_tensor

        if self.ibp:
            u_c_out = op(u_c)
            l_c_out = op(l_c)
        else:
            u_c_out, l_c_out = empty_tensor, empty_tensor

        if self.affine:
            b_u_out = op(b_u)
            b_l_out = op(b_l)

            if len(w_u.shape) == len(b_u.shape):
                w_u_out = op(w_u)
                w_l_out = op(w_l)

            else:

                def step_func(x: tf.Tensor, _: List[tf.Tensor]) -> Tuple[tf.Tensor, List[tf.Tensor]]:
                    return op(x), _

                w_u_out = K.rnn(step_function=step_func, inputs=w_u, initial_states=[], unroll=False)[1]
                w_l_out = K.rnn(step_function=step_func, inputs=w_l, initial_states=[], unroll=False)[1]
        else:
            w_u_out, b_u_out, w_l_out, b_l_out = empty_tensor, empty_tensor, empty_tensor, empty_tensor

        return self.inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
            [x, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out, h_out, g_out]
        )


class DecomonPermute(DecomonLayer, Permute):
    """Forward LiRPA implementation of Reshape layers.
    See Keras official documentation for further details on the Reshape operator
    """

    original_keras_layer_class = Permute

    def __init__(
        self,
        dims: Tuple[int, ...],
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        finetune: bool = False,
        shared: bool = False,
        fast: bool = True,
        **kwargs: Any,
    ):
        """
        Args:
            data_format
            **kwargs
        """
        super().__init__(
            dims=dims,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            mode=mode,
            finetune=finetune,
            shared=shared,
            fast=fast,
            **kwargs,
        )

        if self.mode == ForwardMode.HYBRID:
            self.input_spec = [
                InputSpec(min_ndim=1),  # z
                InputSpec(min_ndim=1),  # u_c
                InputSpec(min_ndim=1),  # w_u
                InputSpec(min_ndim=1),  # b_u
                InputSpec(min_ndim=1),  # l_c
                InputSpec(min_ndim=1),  # w_l
                InputSpec(min_ndim=1),  # b_l
            ]
        elif self.mode == ForwardMode.IBP:
            self.input_spec = [
                InputSpec(min_ndim=1),  # u_c
                InputSpec(min_ndim=1),  # l_c
            ]
        elif self.mode == ForwardMode.AFFINE:
            self.input_spec = [
                InputSpec(min_ndim=1),  # z
                InputSpec(min_ndim=1),  # w_u
                InputSpec(min_ndim=1),  # b_u
                InputSpec(min_ndim=1),  # w_l
                InputSpec(min_ndim=1),  # b_l
            ]
        else:
            raise ValueError(f"Unknown mode {mode}")

        if self.dc_decomp:
            self.input_spec += [InputSpec(min_ndim=1), InputSpec(min_ndim=1)]

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        def op(x: tf.Tensor) -> tf.Tensor:
            return Permute.call(self, x)

        x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = self.inputs_outputs_spec.get_fullinputs_from_inputsformode(
            inputs, compute_ibp_from_affine=False
        )
        dtype = x.dtype
        empty_tensor = self.inputs_outputs_spec.get_empty_tensor(dtype=dtype)

        if self.dc_decomp:
            h_out = op(h)
            g_out = op(g)
        else:
            h_out, g_out = empty_tensor, empty_tensor

        if self.ibp:
            u_c_out = op(u_c)
            l_c_out = op(l_c)
        else:
            u_c_out, l_c_out = empty_tensor, empty_tensor

        if self.affine:
            b_u_out = op(b_u)
            b_l_out = op(b_l)

            if len(w_u.shape) == len(b_u.shape):
                w_u_out = op(w_u)
                w_l_out = op(w_l)
            else:

                def step_func(x: tf.Tensor, _: List[tf.Tensor]) -> Tuple[tf.Tensor, List[tf.Tensor]]:
                    return op(x), _

                w_u_out = K.rnn(step_function=step_func, inputs=w_u, initial_states=[], unroll=False)[1]
                w_l_out = K.rnn(step_function=step_func, inputs=w_l, initial_states=[], unroll=False)[1]
        else:
            w_u_out, b_u_out, w_l_out, b_l_out = empty_tensor, empty_tensor, empty_tensor, empty_tensor

        return self.inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
            [x, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out, h_out, g_out]
        )



from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.layers import Flatten, Layer

from decomon.backward_layers.core import BackwardLayer
from decomon.backward_layers.utils import backward_max_, get_identity_lirpa
from decomon.core import ForwardMode, PerturbationDomain


class BackwardMaxPooling2D(BackwardLayer):
    """Backward  LiRPA of MaxPooling2D"""

    pool_size: Tuple[int, int]
    strides: Tuple[int, int]
    padding: str
    data_format: str
    fast: bool

    def __init__(
        self,
        layer: Layer,
        rec: int = 1,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            layer=layer,
            rec=rec,
            mode=mode,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            **kwargs,
        )
        raise NotImplementedError()

    def _pooling_function_fast(
        self,
        inputs: List[tf.Tensor],
        w_u_out: tf.Tensor,
        b_u_out: tf.Tensor,
        w_l_out: tf.Tensor,
        b_l_out: tf.Tensor,
    ) -> List[tf.Tensor]:
        x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = self.inputs_outputs_spec.get_fullinputs_from_inputsformode(inputs)

        op_flat = Flatten()

        b_u_pooled = K.pool2d(u_c, self.pool_size, self.strides, self.padding, self.data_format, pool_mode="max")
        b_l_pooled = K.pool2d(l_c, self.pool_size, self.strides, self.padding, self.data_format, pool_mode="max")

        b_u_pooled = K.expand_dims(K.expand_dims(op_flat(b_u_pooled), 1), -1)
        b_l_pooled = K.expand_dims(K.expand_dims(op_flat(b_l_pooled), 1), -1)

        y = inputs[-1]
        n_out = w_u_out.shape[-1]

        w_u_out_new = K.concatenate([K.expand_dims(K.expand_dims(0 * (op_flat(y)), 1), -1)] * n_out, -1)
        w_l_out_new = w_u_out_new

        b_u_out_new = (
            K.sum(K.maximum(w_u_out, 0) * b_u_pooled, 2) + K.sum(K.minimum(w_u_out, 0) * b_l_pooled, 2) + b_u_out
        )
        b_l_out_new = (
            K.sum(K.maximum(w_l_out, 0) * b_l_pooled, 2) + K.sum(K.minimum(w_l_out, 0) * b_u_pooled, 2) + b_l_out
        )

        return [w_u_out_new, b_u_out_new, w_l_out_new, b_l_out_new]

    def _pooling_function_not_fast(
        self,
        inputs: List[tf.Tensor],
        w_u_out: tf.Tensor,
        b_u_out: tf.Tensor,
        w_l_out: tf.Tensor,
        b_l_out: tf.Tensor,
    ) -> List[tf.Tensor]:
        """
        Args:
            inputs
            pool_size
            strides
            padding
            data_format

        Returns:

        """
        x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = self.inputs_outputs_spec.get_fullinputs_from_inputsformode(inputs)
        dtype = x.dtype
        empty_tensor = self.inputs_outputs_spec.get_empty_tensor(dtype=dtype)
        y = inputs[-1]
        input_shape = K.int_shape(y)

        if self.data_format in [None, "channels_last"]:
            axis = -1
        else:
            axis = 1

        # initialize vars
        u_c_tmp, w_u_tmp, b_u_tmp, l_c_tmp, w_l_tmp, b_l_tmp = (
            empty_tensor,
            empty_tensor,
            empty_tensor,
            empty_tensor,
            empty_tensor,
            empty_tensor,
        )

        if self.ibp:
            u_c_tmp = K.concatenate([self.internal_op(elem) for elem in tf.split(u_c, input_shape[-1], -1)], -2)
            l_c_tmp = K.concatenate([self.internal_op(elem) for elem in tf.split(l_c, input_shape[-1], -1)], -2)

        if self.affine:
            b_u_tmp = K.concatenate([self.internal_op(elem) for elem in tf.split(b_u, input_shape[-1], -1)], -2)
            b_l_tmp = K.concatenate([self.internal_op(elem) for elem in tf.split(b_l, input_shape[-1], -1)], -2)
            w_u_tmp = K.concatenate([self.internal_op(elem) for elem in tf.split(w_u, input_shape[-1], -1)], -2)
            w_l_tmp = K.concatenate([self.internal_op(elem) for elem in tf.split(w_l, input_shape[-1], -1)], -2)

        if self.dc_decomp:
            raise NotImplementedError()
        else:
            h_tmp, g_tmp = empty_tensor, empty_tensor

        outputs_tmp = self.inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
            [x, u_c_tmp, w_u_tmp, b_u_tmp, l_c_tmp, w_l_tmp, b_l_tmp, h_tmp, g_tmp]
        )

        w_u_out, b_u_out, w_l_out, b_l_out = backward_max_(
            outputs_tmp,
            w_u_out,
            b_u_out,
            w_l_out,
            b_l_out,
            perturbation_domain=self.perturbation_domain,
            mode=self.mode,
            dc_decomp=self.dc_decomp,
            axis=-1,
        )

        # invert the convolution
        op_flat = Flatten()

        # do not do the activation so far
        # get input shape
        input_shape_channelreduced = list(inputs[0].shape[1:])
        n_axis = input_shape_channelreduced[axis]
        input_shape_channelreduced[axis] = 1
        n_dim = np.prod(input_shape_channelreduced)

        # create diagonal matrix
        id_list = [tf.linalg.diag(K.ones_like(op_flat(elem[0][None]))) for elem in tf.split(y, input_shape[axis], axis)]

        id_list = [K.reshape(identity_mat, [-1] + input_shape_channelreduced) for identity_mat in id_list]
        w_list = [self.internal_op(identity_mat) for identity_mat in id_list]

        # flatten
        weights = [K.reshape(op_flat(weights), (n_dim, -1, np.prod(self.pool_size))) for weights in w_list]

        n_0 = weights[0].shape[1]
        n_1 = weights[0].shape[2]

        w_u_out = K.reshape(w_u_out, (-1, 1, n_0, input_shape[axis], w_u_out.shape[-2], n_1))
        w_l_out = K.reshape(w_l_out, (-1, 1, n_0, input_shape[axis], w_l_out.shape[-2], n_1))

        weights = K.expand_dims(K.concatenate([K.expand_dims(K.expand_dims(w, -2), -2) for w in weights], 2), 0)

        w_u_out = K.reshape(
            K.sum(K.expand_dims(w_u_out, 1) * weights, (3, -1)), (-1, 1, n_dim * n_axis, w_u_out.shape[-2])
        )
        w_l_out = K.reshape(
            K.sum(K.expand_dims(w_l_out, 1) * weights, (3, -1)), (-1, 1, n_dim * n_axis, w_l_out.shape[-2])
        )

        return [w_u_out, b_u_out, w_l_out, b_l_out]

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        w_u_out, b_u_out, w_l_out, b_l_out = get_identity_lirpa(inputs)
        if self.fast:
            return self._pooling_function_fast(
                inputs=inputs,
                w_u_out=w_u_out,
                b_u_out=b_u_out,
                w_l_out=w_l_out,
                b_l_out=b_l_out,
            )
        else:
            return self._pooling_function_not_fast(
                inputs=inputs,
                w_u_out=w_u_out,
                b_u_out=b_u_out,
                w_l_out=w_l_out,
                b_l_out=b_l_out,
            )



from typing import Union

import numpy as np
import numpy.typing as npt
import tensorflow as tf
from tensorflow.keras.optimizers import Adam

from decomon.metrics.loss import get_upper_loss
from decomon.models.models import DecomonModel
from decomon.wrapper import get_lower_box, get_upper_box, refine_boxes


#### FORMAL BOUNDS ######
def get_upper_box_tuning(
    model: Union[tf.keras.Model, DecomonModel],
    decomon_model_concat: tf.keras.Model,
    x_min: npt.NDArray[np.float_],
    x_max: npt.NDArray[np.float_],
    batch_size: int = 1,
    n_sub_boxes: int = 1,
    lr: float = 0.1,
    epochs: int = 100,
) -> npt.NDArray[np.float_]:
    """upper bound the maximum of a model in a given box

    Args:
        model: either a Keras model or a Decomon model
        x_min: numpy array for the extremal lower corner of the boxes
        x_max: numpy array for the extremal upper corner of the boxes
        batch_size: for computational efficiency, one can split the
            calls to minibatches
        fast: useful in the forward-backward or in the hybrid-backward
            mode to optimize the scores

    Returns:
        numpy array, vector with upper bounds for adversarial attacks
    """

    if np.min(x_max - x_min) < 0:
        raise UserWarning("Inconsistency Error: x_max < x_min")

    # check that the model is a DecomonModel, else do the conversion
    decomon_model = model

    baseline_upper = get_upper_box(model, x_min, x_max, batch_size=batch_size, n_sub_boxes=n_sub_boxes)

    if n_sub_boxes > 1:
        x_min, x_max = refine_boxes(x_min, x_max, n_sub_boxes)
        # reshape
        shape = list(x_min.shape[2:])
        x_min = np.reshape(x_min, [-1] + shape)
        x_max = np.reshape(x_max, [-1] + shape)

    # reshape x_mmin, x_max
    input_shape = list(decomon_model.input_shape[2:])
    input_dim = np.prod(input_shape)
    x_reshaped = x_min + 0 * x_min
    x_reshaped = x_reshaped.reshape([-1] + input_shape)
    x_min = x_min.reshape((-1, 1, input_dim))
    x_max = x_max.reshape((-1, 1, input_dim))

    z = np.concatenate([x_min, x_max], 1)

    if batch_size > 1:
        # split
        r = 0
        if len(x_reshaped) % batch_size > 0:
            r += 1
        x_min_list = [x_min[batch_size * i : batch_size * (i + 1)] for i in range(len(x_reshaped) // batch_size + r)]
        x_max_list = [x_max[batch_size * i : batch_size * (i + 1)] for i in range(len(x_reshaped) // batch_size + r)]

        results = [
            get_upper_box_tuning(
                decomon_model, decomon_model_concat, x_min_list[i], x_max_list[i], -1, lr=lr, epochs=epochs
            )
            for i in range(len(x_min_list))
        ]

        return np.concatenate(results)

    else:
        # freeze_weights
        model.freeze_weights()

        # create loss
        loss_upper = get_upper_loss(model)
        decomon_model_concat.compile(Adam(lr=lr), loss_upper)

        decomon_model_concat.fit(z, baseline_upper, epochs=epochs, verbose=0)

        upper = get_upper_box(model, x_min, x_max, batch_size=batch_size, n_sub_boxes=n_sub_boxes)

        # reset alpha
        model.reset_finetuning()

        return np.minimum(baseline_upper, upper)


#### FORMAL BOUNDS ######
def get_lower_box_tuning(
    model: Union[tf.keras.Model, DecomonModel],
    decomon_model_concat: tf.keras.Model,
    x_min: npt.NDArray[np.float_],
    x_max: npt.NDArray[np.float_],
    batch_size: int = 1,
    n_sub_boxes: int = 1,
    lr: float = 0.1,
    epochs: int = 100,
) -> npt.NDArray[np.float_]:
    """upper bound the maximum of a model in a given box

    Args:
        model: either a Keras model or a Decomon model
        x_min: numpy array for the extremal lower corner of the boxes
        x_max: numpy array for the extremal upper corner of the boxes
        batch_size: for computational efficiency, one can split the
            calls to minibatches
        fast: useful in the forward-backward or in the hybrid-backward
            mode to optimize the scores

    Returns:
        numpy array, vector with upper bounds for adversarial attacks
    """

    if np.min(x_max - x_min) < 0:
        raise UserWarning("Inconsistency Error: x_max < x_min")

    # check that the model is a DecomonModel, else do the conversion
    decomon_model = model

    baseline_upper = get_lower_box(model, x_min, x_max, batch_size=batch_size, n_sub_boxes=n_sub_boxes)

    if n_sub_boxes > 1:
        x_min, x_max = refine_boxes(x_min, x_max, n_sub_boxes)
        # reshape
        shape = list(x_min.shape[2:])
        x_min = np.reshape(x_min, [-1] + shape)
        x_max = np.reshape(x_max, [-1] + shape)

    # reshape x_mmin, x_max
    input_shape = list(decomon_model.input_shape[2:])
    input_dim = np.prod(input_shape)
    x_reshaped = x_min + 0 * x_min
    x_reshaped = x_reshaped.reshape([-1] + input_shape)
    x_min = x_min.reshape((-1, 1, input_dim))
    x_max = x_max.reshape((-1, 1, input_dim))

    z = np.concatenate([x_min, x_max], 1)

    if batch_size > 0 and batch_size != len(x_min):
        # split
        r = 0
        if len(x_reshaped) % batch_size > 0:
            r += 1
        x_min_list = [x_min[batch_size * i : batch_size * (i + 1)] for i in range(len(x_reshaped) // batch_size + r)]
        x_max_list = [x_max[batch_size * i : batch_size * (i + 1)] for i in range(len(x_reshaped) // batch_size + r)]

        results = [
            get_lower_box_tuning(
                decomon_model, decomon_model_concat, x_min_list[i], x_max_list[i], -1, lr=lr, epochs=epochs
            )
            for i in range(len(x_min_list))
        ]

        model.reset_finetuning()

        return np.concatenate(results)
    else:
        # freeze_weights
        model.freeze_weights()

        # create loss
        loss_upper = get_upper_loss(model)
        decomon_model_concat.compile(Adam(lr=lr), loss_upper)

        decomon_model_concat.fit(z, baseline_upper, epochs=epochs, verbose=0)

        upper = get_lower_box(model, x_min, x_max, batch_size=batch_size, n_sub_boxes=n_sub_boxes)

        # reset alpha
        model.reset_finetuning()

        return np.minimum(baseline_upper, upper)



from decomon.core import ForwardMode
from decomon.metrics.loss import DecomonLossFusion, DecomonRadiusRobust
from decomon.metrics.metric import (
    AdversarialCheck,
    AdversarialScore,
    MetricMode,
    UpperScore,
)


def test_decomon_loss():
    backward = True
    mode = ForwardMode.AFFINE

    layer = DecomonLossFusion(mode=mode, backward=backward)
    config = layer.get_config()
    assert config["backward"] == backward
    assert config["mode"] == mode

    layer = DecomonRadiusRobust(backward=backward, mode=mode)
    config = layer.get_config()
    assert config["backward"] == backward
    assert config["mode"] == mode


def test_metric():
    ibp, affine, mode, perturbation_domain = True, False, MetricMode.BACKWARD, {}
    for cls in [AdversarialCheck, AdversarialScore, UpperScore]:
        layer = cls(ibp=ibp, affine=affine, mode=mode, perturbation_domain=perturbation_domain)
        config = layer.get_config()
        assert config["ibp"] == ibp
        assert config["affine"] == affine
        assert config["mode"] == mode
        assert "perturbation_domain" in config



import warnings

import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.layers import Conv2D, Input


def get_toeplitz(conv_layer: Conv2D, flatten: bool = True) -> tf.Tensor:
    """Express formally the affine component of the convolution
    Conv is a linear operator but its affine component is implicit
    we use im2col and extract_patches to express the affine matrix
    Note that this matrix is Toeplitz

    Args:
        conv_layer: Keras Conv2D layer or Decomon Conv2D layer
        flatten (optional): convert the affine component as a 2D matrix (n_in, n_out). Defaults to True.

    Returns:
         the affine operator W: conv(x)= Wx + bias
    """
    if conv_layer.dtype == "float16":
        warnings.warn("Loss of precision for float16 !")
    if conv_layer.data_format == "channels_last":
        return get_toeplitz_channels_last(conv_layer, flatten)
    else:
        return get_toeplitz_channels_first(conv_layer, flatten)


def get_toeplitz_channels_last(conv_layer: Conv2D, flatten: bool = True) -> tf.Tensor:
    """Express formally the affine component of the convolution for data_format=channels_last
    Conv is a linear operator but its affine component is implicit
    we use im2col and extract_patches to express the affine matrix
    Note that this matrix is Toeplitz

    Args:
        conv_layer: Keras Conv2D layer or Decomon Conv2D layer
        flatten (optional): convert the affine component as a 2D matrix (n_in, n_out). Defaults to True.

    Returns:
         the affine operator W: conv(x)= Wx + bias
    """

    input_shape = conv_layer.get_input_shape_at(0)
    if isinstance(input_shape, list):
        input_shape = input_shape[-1]
    _, w_in, h_in, c_in = input_shape
    padding = conv_layer.padding.upper()
    output_shape = conv_layer.get_output_shape_at(0)
    if isinstance(output_shape, list):
        output_shape = output_shape[-1]
    _, w_out, h_out, c_out = output_shape

    kernel_filter = conv_layer.kernel
    filter_size = kernel_filter.shape[0]
    stride_rows, stride_cols = conv_layer.strides
    rates_rows, rates_cols = conv_layer.dilation_rate

    diag = tf.linalg.diag(tf.ones((1, w_in * h_in * c_in)))
    diag_ = tf.reshape(diag, (w_in * h_in * c_in, w_in, h_in, c_in))

    diag_patches = tf.image.extract_patches(
        diag_,
        [1, filter_size, filter_size, 1],
        strides=[1, stride_rows, stride_cols, 1],
        rates=[1, rates_rows, rates_cols, 1],
        padding=padding,
    )

    diag_patches_ = tf.reshape(diag_patches, (w_in, h_in, c_in, w_out, h_out, filter_size**2, c_in))

    shape = np.arange(len(diag_patches_.shape))
    shape[-1] -= 1
    shape[-2] += 1
    diag_patches_ = K.permute_dimensions(diag_patches_, shape)
    kernel = conv_layer.kernel  # (filter_size, filter_size, c_in, c_out)

    kernel = K.permute_dimensions(tf.reshape(kernel, (filter_size**2, c_in, c_out)), (1, 2, 0))[
        None, None, None, None, None
    ]  # (1,1,c_in, c_out, filter_size^2)

    # element-wise multiplication is only compatible with float32
    diag_patches_ = K.cast(diag_patches_, "float32")
    kernel = K.cast(kernel, "float32")

    w = K.sum(K.expand_dims(diag_patches_, -2) * kernel, (5, 7))
    # kernel shape: (1,1,1,1,1,c_in, c_out, filter_size^2)
    # diag_patches shape: (w_in, h_in, c_in, w_out, h_out, c_in, 1,     filter_size^2)
    # (w_in, h_in, c_in, w_out, h_out, c_in, c_out, filter^2)

    w = K.cast(w, conv_layer.dtype)
    # cast w for dtype

    if flatten:
        return K.reshape(w, (w_in * h_in * c_in, w_out * h_out * c_out))
    else:
        return w


def get_toeplitz_channels_first(conv_layer: Conv2D, flatten: bool = True) -> tf.Tensor:
    """Express formally the affine component of the convolution for data_format=channels_first
    Conv is a linear operator but its affine component is implicit
    we use im2col and extract_patches to express the affine matrix
    Note that this matrix is Toeplitz

    Args:
        conv_layer: Keras Conv2D layer or Decomon Conv2D layer
        flatten (optional): convert the affine component as a 2D matrix (n_in, n_out). Defaults to True.

    Returns:
        the affine operator W: conv(x)= Wx + bias
    """

    input_shape = conv_layer.get_input_shape_at(0)
    if isinstance(input_shape, list):
        input_shape = input_shape[-1]
    _, c_in, w_in, h_in = input_shape
    padding = conv_layer.padding.upper()
    output_shape = conv_layer.get_output_shape_at(0)
    if isinstance(output_shape, list):
        output_shape = output_shape[-1]
    _, c_out, w_out, h_out = output_shape
    kernel_filter = conv_layer.kernel
    filter_size = kernel_filter.shape[0]

    diag = tf.linalg.diag(tf.ones((1, w_in * h_in * c_in)))
    diag_ = tf.reshape(diag, (w_in * h_in * c_in, w_in, h_in, c_in))

    diag_patches = tf.image.extract_patches(
        diag_, [1, filter_size, filter_size, 1], [1, 1, 1, 1], [1, 1, 1, 1], padding=padding
    )

    diag_patches_ = tf.reshape(diag_patches, (w_in, h_in, c_in, w_out, h_out, filter_size**2, c_in))

    shape = np.arange(len(diag_patches_.shape))
    shape[-1] -= 1
    shape[-2] += 1
    diag_patches_ = K.permute_dimensions(diag_patches_, shape)
    kernel = conv_layer.kernel  # (filter_size, filter_size, c_in, c_out)

    kernel = K.permute_dimensions(tf.reshape(kernel, (filter_size**2, c_in, c_out)), (1, 2, 0))[
        None, None, None, None, None
    ]  # (1,1,c_in, c_out, filter_size^2)

    w = K.sum(K.expand_dims(diag_patches_, -2) * kernel, (5, 7))
    # kernel shape: (1,      1,     1,   1,    1,    c_in, c_out, filter_size^2)
    # diag_patches shape: (w_in, h_in, c_in, w_out, h_out, c_in, 1,     filter_size^2)
    # (w_in, h_in, c_in, w_out, h_out, c_in, c_out, filter^2)

    w = K.cast(w, conv_layer.dtype)
    # cast w for dtype

    if flatten:
        return K.reshape(w, (w_in * h_in * c_in, w_out * h_out * c_out))
    else:
        return w



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
from subprocess import DEVNULL, CalledProcessError, check_output

import decomon

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
project = "decomon"
copyright = "2022, Airbus"
author = "Airbus"

DEV_VERSION_PATTERN = "dev"
GIT_SYMREF_CMD = ["git", "symbolic-ref", "--short", "HEAD"]

version = decomon.__version__
if DEV_VERSION_PATTERN in version:
    # not a release but rather dev doc => we try to use the current branch name instead of version 0.0.0
    try:
        version = check_output(GIT_SYMREF_CMD, stderr=DEVNULL, encoding="utf-8").strip()
    except CalledProcessError:
        pass
release = version

master_doc = "index"
language = "en"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
]

templates_path = ["_templates"]
exclude_patterns = ["*.template.*"]

# myST configuration
myst_enable_extensions = [
    "dollarmath",
    "amsmath",
    "deflist",
    "fieldlist",
    "html_admonition",
    "html_image",
    "colon_fence",
    "smartquotes",
    "replacements",
    "strikethrough",
    "substitution",
    "tasklist",
]
myst_number_code_blocks = ["typescript"]
myst_heading_anchors = 3  # up to h3 titles will have autogenerated anchors
myst_footnote_transition = True
myst_dmath_double_inline = True


# autosummary config
autosummary_generate = True

# napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = False


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
html_show_sourcelink = False
html_theme_options = {
    "style_external_links": True,
}

html_js_files = ["versions.js"]
AUTODOC_SHOW_VERSIONS = os.environ.get("AUTODOC_SHOW_VERSIONS", "True").lower() != "false"
html_context = {
    "AUTODOC_SHOW_VERSIONS": AUTODOC_SHOW_VERSIONS,
    "current_version": version,
}



import warnings
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.layers import Layer

from decomon.core import (
    BoxDomain,
    ForwardMode,
    InputsOutputsSpec,
    PerturbationDomain,
    Slope,
)
from decomon.utils import (
    get_linear_hull_relu,
    get_linear_hull_s_shape,
    get_linear_hull_sigmoid,
    get_linear_hull_tanh,
    get_linear_softplus_hull,
    sigmoid_prime,
    softsign_prime,
    tanh_prime,
)

ELU = "elu"
SELU = "selu"
SOFTPLUS = "softplus"
SOFTSIGN = "softsign"
SOFTMAX = "softmax"
RELU_ = "relu_"
RELU = "relu"
SIGMOID = "sigmoid"
TANH = "tanh"
EXPONENTIAL = "exponential"
HARD_SIGMOID = "hard_sigmoid"
LINEAR = "linear"


def backward_relu(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    alpha: float = 0.0,
    max_value: Optional[float] = None,
    threshold: float = 0.0,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """Backward  LiRPA of relu

    Args:
        inputs
        dc_decomp
        perturbation_domain
        alpha
        max_value
        threshold
        slope
        mode

    Returns:

    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    mode = ForwardMode(mode)
    if dc_decomp:
        raise NotImplementedError()

    if threshold != 0:
        raise NotImplementedError()

    if not alpha and max_value is None:
        # default values: return relu_(x) = max(x, 0)
        inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)
        x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(inputs)
        input_shape = inputs_outputs_spec.get_input_shape(inputs)
        bounds = get_linear_hull_relu(upper=u_c, lower=l_c, slope=slope, **kwargs)
        dim = np.prod(input_shape[1:])
        return [K.reshape(elem, (-1, dim)) for elem in bounds]

    raise NotImplementedError()


def backward_sigmoid(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """Backward  LiRPA of sigmoid

    Args:
        inputs
        dc_decomp
        perturbation_domain
        slope
        mode

    Returns:

    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    if dc_decomp:
        raise NotImplementedError()
    mode = ForwardMode(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)
    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(inputs)
    return get_linear_hull_sigmoid(u_c, l_c, slope=slope, **kwargs)


def backward_tanh(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """Backward  LiRPA of tanh

    Args:
        inputs
        dc_decomp
        perturbation_domain
        slope
        mode

    Returns:

    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    if dc_decomp:
        raise NotImplementedError()
    mode = ForwardMode(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)
    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(inputs)
    return get_linear_hull_tanh(u_c, l_c, slope=slope, **kwargs)


def backward_hard_sigmoid(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """Backward  LiRPA of hard sigmoid

    Args:
        inputs
        dc_decomp
        perturbation_domain
        slope
        mode

    Returns:

    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    if dc_decomp:
        raise NotImplementedError()
    mode = ForwardMode(mode)
    raise NotImplementedError()


def backward_elu(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """Backward  LiRPA of Exponential Linear Unit

    Args:
        inputs
        dc_decomp
        perturbation_domain
        slope
        mode

    Returns:

    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    if dc_decomp:
        raise NotImplementedError()
    mode = ForwardMode(mode)
    slope = Slope(slope)
    raise NotImplementedError()


def backward_selu(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """Backward LiRPA of Scaled Exponential Linear Unit (SELU)

    Args:
        inputs
        dc_decomp
        perturbation_domain
        slope
        mode

    Returns:

    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    if dc_decomp:
        raise NotImplementedError()
    mode = ForwardMode(mode)
    slope = Slope(slope)
    raise NotImplementedError()


def backward_linear(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """Backward LiRPA of linear

    Args:
        inputs
        dc_decomp
        perturbation_domain
        slope
        mode

    Returns:

    """
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    mode = ForwardMode(mode)
    slope = Slope(slope)
    raise NotImplementedError()


def backward_exponential(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """Backward LiRPAof exponential

    Args:
        inputs
        dc_decomp
        perturbation_domain
        slope
        mode

    Returns:

    """
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    if dc_decomp:
        raise NotImplementedError()
    mode = ForwardMode(mode)
    slope = Slope(slope)
    raise NotImplementedError()


def backward_softplus(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """Backward LiRPA of softplus

    Args:
        inputs
        dc_decomp
        perturbation_domain
        slope
        mode

    Returns:

    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    if dc_decomp:
        raise NotImplementedError()
    mode = ForwardMode(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)
    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(inputs)
    return get_linear_softplus_hull(u_c, l_c, slope=slope, **kwargs)


def backward_softsign(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """Backward LiRPA of softsign

    Args:
        inputs
        w_u_out
        b_u_out
        w_l_out
        b_l_out
        perturbation_domain
        slope: backward slope
        mode

    Returns:

    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()

    bounds = get_linear_hull_s_shape(
        inputs, func=K.softsign, f_prime=softsign_prime, perturbation_domain=perturbation_domain, mode=mode
    )
    shape = np.prod(inputs[-1].shape[1:])
    return [K.reshape(elem, (-1, shape)) for elem in bounds]


def backward_softsign_(
    inputs: List[tf.Tensor],
    w_u_out: tf.Tensor,
    b_u_out: tf.Tensor,
    w_l_out: tf.Tensor,
    b_l_out: tf.Tensor,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    **kwargs: Any,
) -> List[tf.Tensor]:
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    w_u_0, b_u_0, w_l_0, b_l_0 = get_linear_hull_s_shape(
        inputs, func=K.softsign, f_prime=softsign_prime, perturbation_domain=perturbation_domain, mode=mode, slope=slope
    )

    w_u_0 = K.expand_dims(w_u_0, -1)
    w_l_0 = K.expand_dims(w_l_0, -1)

    b_u_0 = K.expand_dims(b_u_0, -1)
    b_l_0 = K.expand_dims(b_l_0, -1)

    z_value = K.cast(0.0, dtype=inputs[0].dtype)
    b_u_out = K.sum(K.maximum(z_value, w_u_out) * b_u_0 + K.minimum(z_value, w_u_out) * b_l_0, 2) + b_u_out
    b_l_out = K.sum(K.maximum(z_value, w_l_out) * b_l_0 + K.minimum(z_value, w_l_out) * b_u_0, 2) + b_l_out
    w_u_out = K.maximum(z_value, w_u_out) * w_u_0 + K.minimum(z_value, w_u_out) * w_l_0
    w_l_out = K.maximum(z_value, w_l_out) * w_l_0 + K.minimum(z_value, w_l_out) * w_u_0

    return [w_u_out, b_u_out, w_l_out, b_l_out]


def backward_softmax(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    axis: int = -1,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """Backward LiRPA of softmax

    Args:
        inputs
        dc_decomp
        perturbation_domain
        slope
        mode
        axis

    Returns:

    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    if dc_decomp:
        raise NotImplementedError()
    mode = ForwardMode(mode)
    slope = Slope(slope)
    raise NotImplementedError()


def deserialize(name: str) -> Callable[..., List[tf.Tensor]]:
    """Get the activation from name.

    Args:
        name: name of the method.
    among the implemented Keras activation function.

    Returns:

    """
    name = name.lower()

    if name == SOFTMAX:
        return backward_softmax
    if name == ELU:
        return backward_elu
    if name == SELU:
        return backward_selu
    if name == SOFTPLUS:
        return backward_softplus
    if name == SOFTSIGN:
        return backward_softsign
    if name == SIGMOID:
        return backward_sigmoid
    if name == TANH:
        return backward_tanh
    if name in [RELU, RELU_]:
        return backward_relu
    if name == EXPONENTIAL:
        return backward_exponential
    if name == LINEAR:
        return backward_linear
    raise ValueError("Could not interpret " "activation function identifier:", name)


def get(identifier: Any) -> Callable[..., List[tf.Tensor]]:
    """Get the `identifier` activation function.

    Args:
        identifier: None or str, name of the function.

    Returns:
        The activation function, `linear` if `identifier` is None.

    """
    if identifier is None:
        return backward_linear
    if isinstance(identifier, str):
        return deserialize(identifier)
    elif callable(identifier):
        if isinstance(identifier, Layer):
            warnings.warn(
                "Do not pass a layer instance (such as {identifier}) as the "
                "activation argument of another layer. Instead, advanced "
                "activation layers should be used just like any other "
                "layer in a model.".format(identifier=identifier.__class__.__name__)
            )
        return identifier
    else:
        raise ValueError("Could not interpret " "activation function identifier:", identifier)



# Test unit for decomon with Dense layers


import numpy as np
import pytest
import tensorflow.keras.backend as K
from tensorflow.keras.layers import (
    Add,
    Average,
    Concatenate,
    Maximum,
    Minimum,
    Multiply,
    Subtract,
)
from tensorflow.keras.models import Model

from decomon.core import ForwardMode, get_affine, get_ibp
from decomon.layers.convert import to_decomon
from decomon.layers.decomon_merge_layers import (
    DecomonAdd,
    DecomonAverage,
    DecomonConcatenate,
    DecomonMaximum,
    DecomonMinimum,
    DecomonMultiply,
    DecomonSubtract,
)


def add_op(x, y):
    return x + y


def subtract_op(x, y):
    return x - y


def multiply_op(x, y):
    return x * y


def average_op(x, y):
    return (x + y) / 2.0


def concatenate_op(x, y):
    return np.concatenate([x, y], -1)


@pytest.mark.parametrize(
    "decomon_op_class, tensor_op, decomon_op_kwargs",
    [
        (DecomonAdd, add_op, {}),
        (DecomonSubtract, subtract_op, {}),
        (DecomonAverage, average_op, {}),
        (DecomonMaximum, np.maximum, {}),
        (DecomonMinimum, np.minimum, {}),
        (DecomonConcatenate, concatenate_op, {"axis": -1}),
        (DecomonMultiply, multiply_op, {}),
    ],
)
def test_DecomonOp_1D_box(decomon_op_class, tensor_op, decomon_op_kwargs, n, mode, floatx, decimal, helpers):
    dc_decomp = False

    #  tensor inputs
    inputs_0 = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_1 = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_for_mode_0 = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs_0, mode=mode, dc_decomp=dc_decomp)
    inputs_for_mode_1 = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs_1, mode=mode, dc_decomp=dc_decomp)

    # numpy inputs
    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=dc_decomp)
    input_ref_ = helpers.get_input_ref_from_full_inputs(inputs=inputs_)

    # original output
    output_ref_ = tensor_op(input_ref_, input_ref_)

    # decomon output
    decomon_op = decomon_op_class(dc_decomp=dc_decomp, mode=mode, dtype=K.floatx(), **decomon_op_kwargs)
    output = decomon_op(inputs_for_mode_0 + inputs_for_mode_1)
    f_decomon = K.function(inputs_0 + inputs_1, output)
    outputs_ = f_decomon(inputs_ + inputs_)

    #  check bounds consistency
    helpers.assert_decomon_layer_output_properties_box(
        full_inputs=inputs_,
        output_ref=output_ref_,
        outputs_for_mode=outputs_,
        mode=mode,
        dc_decomp=dc_decomp,
        decimal=decimal,
    )


@pytest.mark.parametrize(
    "decomon_op_class, tensor_op, decomon_op_kwargs",
    [
        (DecomonAdd, add_op, {}),
        (DecomonSubtract, subtract_op, {}),
        (DecomonAverage, average_op, {}),
        (DecomonMaximum, np.maximum, {}),
        (DecomonMinimum, np.minimum, {}),
        (DecomonConcatenate, concatenate_op, {"axis": -1}),
        (DecomonMultiply, multiply_op, {}),
    ],
)
def test_DecomonOp_multiD_box(decomon_op_class, tensor_op, decomon_op_kwargs, odd, mode, floatx, decimal, helpers):
    dc_decomp = False

    #  tensor inputs
    inputs_0 = helpers.get_tensor_decomposition_multid_box(odd, dc_decomp=dc_decomp)
    inputs_1 = helpers.get_tensor_decomposition_multid_box(odd, dc_decomp=dc_decomp)
    inputs_for_mode_0 = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs_0, mode=mode, dc_decomp=dc_decomp)
    inputs_for_mode_1 = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs_1, mode=mode, dc_decomp=dc_decomp)

    # numpy inputs
    inputs_ = helpers.get_standard_values_multid_box(odd, dc_decomp=dc_decomp)
    input_ref_ = helpers.get_input_ref_from_full_inputs(inputs=inputs_)

    # original output
    output_ref_ = tensor_op(input_ref_, input_ref_)

    # decomon output
    decomon_op = decomon_op_class(dc_decomp=dc_decomp, mode=mode, dtype=K.floatx(), **decomon_op_kwargs)
    output = decomon_op(inputs_for_mode_0 + inputs_for_mode_1)
    f_decomon = K.function(inputs_0 + inputs_1, output)
    outputs_ = f_decomon(inputs_ + inputs_)

    #  check bounds consistency
    helpers.assert_decomon_layer_output_properties_box(
        full_inputs=inputs_,
        output_ref=output_ref_,
        outputs_for_mode=outputs_,
        mode=mode,
        dc_decomp=dc_decomp,
        decimal=decimal,
    )


@pytest.mark.parametrize(
    "layer_class, tensor_op, layer_kwargs",
    [
        (Add, add_op, {}),
        (Average, average_op, {}),
        (Subtract, subtract_op, {}),
        (Maximum, np.maximum, {}),
        (Minimum, np.minimum, {}),
        (Concatenate, concatenate_op, {"axis": -1}),
        (Multiply, multiply_op, {}),
    ],
)
def test_Decomon_1D_box_to_decomon(layer_class, tensor_op, layer_kwargs, n, helpers):
    dc_decomp = False
    mode = ForwardMode.HYBRID
    ibp = get_ibp(mode=mode)
    affine = get_affine(mode=mode)
    decimal = 5

    #  tensor inputs
    inputs_0 = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_1 = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_for_mode_0 = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs_0, mode=mode, dc_decomp=dc_decomp)
    inputs_for_mode_1 = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs_1, mode=mode, dc_decomp=dc_decomp)
    input_ref_0 = helpers.get_input_ref_from_full_inputs(inputs=inputs_0)
    input_ref_1 = helpers.get_input_ref_from_full_inputs(inputs=inputs_1)

    # numpy inputs
    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=dc_decomp)
    input_ref_ = helpers.get_input_ref_from_full_inputs(inputs=inputs_)

    # original output
    output_ref_ = tensor_op(input_ref_, input_ref_)

    # to_decomon
    ref_op = layer_class(dtype=K.floatx(), **layer_kwargs)
    ref_op([input_ref_0, input_ref_1])
    decomon_op = to_decomon(
        ref_op, input_dim=helpers.get_input_dim_from_full_inputs(inputs_0), dc_decomp=dc_decomp, affine=affine, ibp=ibp
    )

    # decomon output
    output = decomon_op(inputs_for_mode_0 + inputs_for_mode_1)
    f_decomon = K.function(inputs_0 + inputs_1, output)
    outputs_ = f_decomon(inputs_ + inputs_)

    #  check bounds consistency
    helpers.assert_decomon_layer_output_properties_box_linear(
        full_inputs=inputs_,
        output_ref=output_ref_,
        outputs_for_mode=outputs_,
        mode=mode,
        dc_decomp=dc_decomp,
        decimal=decimal,
    )


#### to_decomon multiD


@pytest.mark.parametrize(
    "layer_class, tensor_op, layer_kwargs",
    [
        (Add, add_op, {}),
        (Average, average_op, {}),
        (Subtract, subtract_op, {}),
        (Maximum, np.maximum, {}),
        (Minimum, np.minimum, {}),
        (Concatenate, concatenate_op, {"axis": -1}),
        (Multiply, multiply_op, {}),
    ],
)
def test_Decomon_multiD_box_to_decomon(layer_class, tensor_op, layer_kwargs, odd, helpers):
    dc_decomp = False
    mode = ForwardMode.HYBRID
    ibp = get_ibp(mode=mode)
    affine = get_affine(mode=mode)
    decimal = 5

    #  tensor inputs
    inputs_0 = helpers.get_tensor_decomposition_multid_box(odd, dc_decomp=dc_decomp)
    inputs_1 = helpers.get_tensor_decomposition_multid_box(odd, dc_decomp=dc_decomp)
    inputs_for_mode_0 = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs_0, mode=mode, dc_decomp=dc_decomp)
    inputs_for_mode_1 = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs_1, mode=mode, dc_decomp=dc_decomp)
    input_ref_0 = helpers.get_input_ref_from_full_inputs(inputs=inputs_0)
    input_ref_1 = helpers.get_input_ref_from_full_inputs(inputs=inputs_1)

    # numpy inputs
    inputs_ = helpers.get_standard_values_multid_box(odd, dc_decomp=dc_decomp)
    input_ref_ = helpers.get_input_ref_from_full_inputs(inputs=inputs_)

    # original output
    output_ref_ = tensor_op(input_ref_, input_ref_)

    #  to_decomon
    ref_op = layer_class(dtype=K.floatx(), **layer_kwargs)
    ref_op([input_ref_0, input_ref_1])
    decomon_op = to_decomon(
        ref_op, input_dim=helpers.get_input_dim_from_full_inputs(inputs_0), dc_decomp=dc_decomp, affine=affine, ibp=ibp
    )

    # decomon output
    output = decomon_op(inputs_for_mode_0 + inputs_for_mode_1)
    f_decomon = K.function(inputs_0 + inputs_1, output)
    outputs_ = f_decomon(inputs_ + inputs_)

    #  check bounds consistency
    helpers.assert_decomon_layer_output_properties_box_linear(
        full_inputs=inputs_,
        output_ref=output_ref_,
        outputs_for_mode=outputs_,
        mode=mode,
        dc_decomp=dc_decomp,
        decimal=decimal,
    )



import tensorflow.keras.backend as K
from tensorflow.keras.layers import Dense, Input

from decomon.backward_layers.convert import to_backward
from decomon.core import ForwardMode
from decomon.layers.decomon_layers import DecomonDense


def test_Backward_Dense_1D_box(n, use_bias, mode, floatx, decimal, helpers):
    dc_decomp = False

    #  tensor inputs
    inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    input_dim = helpers.get_input_dim_from_full_inputs(inputs)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs)

    # numpy inputs
    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=dc_decomp)

    # keras layer
    keras_layer = Dense(1, use_bias=use_bias, dtype=K.floatx())
    keras_layer(input_ref)  # init weights

    # get backward layer
    backward_layer = to_backward(keras_layer, mode=mode, input_dim=input_dim)

    # backward outputs
    outputs = backward_layer(inputs_for_mode)
    f_decomon = K.function(inputs, outputs)
    outputs_ = f_decomon(inputs_)

    # check bounds consistency
    helpers.assert_backward_layer_output_properties_box_linear(
        full_inputs=inputs_,
        backward_outputs=outputs_,
        decimal=decimal,
    )


def test_Backward_Dense_multiD_box(odd, floatx, decimal, mode, helpers):
    dc_decomp = False
    use_bias = True

    #  tensor inputs
    inputs = helpers.get_tensor_decomposition_multid_box(odd=odd, dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    input_dim = helpers.get_input_dim_from_full_inputs(inputs)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs)

    # numpy inputs
    inputs_ = helpers.get_standard_values_multid_box(odd, dc_decomp=dc_decomp)

    # keras layer
    keras_layer = Dense(1, use_bias=use_bias, dtype=K.floatx())
    keras_layer(input_ref)  # init weights

    # get backward layer
    backward_layer = to_backward(keras_layer, mode=mode, input_dim=input_dim)

    # backward outputs
    outputs = backward_layer(inputs_for_mode)
    f_decomon = K.function(inputs, outputs)
    outputs_ = f_decomon(inputs_)

    # check bounds consistency
    helpers.assert_backward_layer_output_properties_box_linear(
        full_inputs=inputs_,
        backward_outputs=outputs_,
        decimal=decimal,
    )


def test_Backward_DecomonDense_1D_box(n, helpers):
    dc_decomp = False
    use_bias = True
    mode = ForwardMode.HYBRID
    decimal = 5

    #  tensor inputs
    inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    input_dim = helpers.get_input_dim_from_full_inputs(inputs)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs)

    # numpy inputs
    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=dc_decomp)

    # decomon layer
    decomon_layer = DecomonDense(1, use_bias=use_bias, dc_decomp=dc_decomp, mode=mode, dtype=K.floatx())
    decomon_layer(inputs_for_mode)  # init weights

    # get backward layer
    backward_layer = to_backward(decomon_layer, input_dim=input_dim)

    # backward outputs
    outputs = backward_layer(inputs_for_mode)
    f_decomon = K.function(inputs, outputs)
    outputs_ = f_decomon(inputs_)

    # check bounds consistency
    helpers.assert_backward_layer_output_properties_box_linear(
        full_inputs=inputs_,
        backward_outputs=outputs_,
        decimal=decimal,
    )

    # try to call the backward layer with previous bounds
    w_out = Input((1, 1))
    b_out = Input((1,))
    previous_bounds = [w_out, b_out, w_out, b_out]
    backward_layer(inputs_for_mode + previous_bounds)



from typing import Any, Dict, List, Optional, Union

import tensorflow as tf

from decomon.core import BoxDomain, ForwardMode, PerturbationDomain
from decomon.layers.utils import exp, expand_dims, log, sum
from decomon.utils import add, minus

# compute categorical cross entropy


def categorical_cross_entropy(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    perturbation_domain: Optional[PerturbationDomain] = None,
) -> List[tf.Tensor]:
    # step 1: exponential
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    outputs = exp(inputs, mode=mode, perturbation_domain=perturbation_domain, dc_decomp=dc_decomp)
    # step 2: sum
    outputs = sum(outputs, axis=-1, mode=mode)
    # step 3: log
    outputs = log(outputs, dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)
    outputs = expand_dims(outputs, dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain, axis=-1)
    # step 4: - inputs
    return add(
        minus(inputs, mode=mode, perturbation_domain=perturbation_domain, dc_decomp=dc_decomp),
        outputs,
        mode=mode,
        perturbation_domain=perturbation_domain,
        dc_decomp=dc_decomp,
    )



from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import tensorflow as tf
import tensorflow.keras as keras
from tensorflow.keras import backend as K
from tensorflow.keras.constraints import Constraint
from tensorflow.keras.initializers import Initializer
from tensorflow.keras.layers import Layer

from decomon.core import (
    BallDomain,
    BoxDomain,
    ForwardMode,
    InputsOutputsSpec,
    PerturbationDomain,
    Slope,
    get_affine,
    get_ibp,
)
from decomon.utils import add, get_linear_softplus_hull, maximum, minimum, minus, relu_


def is_a_merge_layer(layer: Layer) -> bool:
    return hasattr(layer, "_merge_function")


class NonPos(Constraint):
    """Constrains the weights to be non-negative."""

    def __call__(self, w: tf.Tensor) -> tf.Tensor:
        return K.minimum(w, 0.0)


class NonNeg(Constraint):
    """Constrains the weights to be non-negative."""

    def __call__(self, w: tf.Tensor) -> tf.Tensor:
        return K.maximum(w, 0.0)


class ClipAlpha(Constraint):
    """Cosntraints the weights to be between 0 and 1."""

    def __call__(self, w: tf.Tensor) -> tf.Tensor:
        return K.clip(w, 0.0, 1.0)


class ClipAlphaGrid(Constraint):
    """Cosntraints the weights to be between 0 and 1."""

    def __call__(self, w: tf.Tensor) -> tf.Tensor:
        w = K.clip(w, 0.0, 1.0)
        w /= K.maximum(K.sum(w, 0), 1.0)[None]
        return w


class ClipAlphaAndSumtoOne(Constraint):
    """Cosntraints the weights to be between 0 and 1."""

    def __call__(self, w: tf.Tensor) -> tf.Tensor:
        w = K.clip(w, 0.0, 1.0)
        # normalize the first colum to 1
        w_scale = K.maximum(K.sum(w, 0), K.epsilon())
        return w / w_scale[:, None, None, None]


class MultipleConstraint(Constraint):
    """stacking multiple constraints"""

    def __init__(self, constraint_0: Optional[Constraint], constraint_1: Constraint, **kwargs: Any):
        super().__init__(**kwargs)
        if constraint_0:
            self.constraints = [constraint_0, constraint_1]
        else:
            self.constraints = [constraint_1]

    def __call__(self, w: tf.Tensor) -> tf.Tensor:
        for c in self.constraints:
            w = c.__call__(w)

        return w


class Project_initializer_pos(Initializer):
    """Initializer that generates tensors initialized to 1."""

    def __init__(self, initializer: Initializer, **kwargs: Any):
        super().__init__(**kwargs)
        self.initializer = initializer

    def __call__(self, shape: tf.TensorShape, dtype: Optional[tf.DType] = None, **kwargs: Any) -> tf.Tensor:
        w = self.initializer.__call__(shape, dtype)
        return K.maximum(0.0, w)


class Project_initializer_neg(Initializer):
    """Initializer that generates tensors initialized to 1."""

    def __init__(self, initializer: Initializer, **kwargs: Any):
        super().__init__(**kwargs)
        self.initializer = initializer

    def __call__(self, shape: tf.TensorShape, dtype: Optional[tf.DType] = None, **kwargs: Any) -> tf.Tensor:
        w = self.initializer.__call__(shape, dtype)
        return K.minimum(0.0, w)


def softplus_(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    **kwargs: Any,
) -> List[tf.Tensor]:
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    mode = ForwardMode(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)

    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(inputs)
    dtype = x.dtype
    empty_tensor = inputs_outputs_spec.get_empty_tensor(dtype=dtype)

    u_c_out = K.softplus(u_c)
    l_c_out = K.softplus(l_c)

    if mode in [ForwardMode.AFFINE, ForwardMode.HYBRID]:
        w_u_out, b_u_out, w_l_out, b_l_out = get_linear_softplus_hull(upper=u_c, lower=l_c, slope=slope, **kwargs)
        b_u_out = w_u_out * b_u + b_u_out
        b_l_out = w_l_out * b_l + b_l_out
        w_u_out = K.expand_dims(w_u_out, 1) * w_u
        w_l_out = K.expand_dims(w_l_out, 1) * w_l
    else:
        w_u_out, b_u_out, w_l_out, b_l_out = empty_tensor, empty_tensor, empty_tensor, empty_tensor

    if dc_decomp:
        raise NotImplementedError("Not yet implemented for dc_decomp=True")
    else:
        h_out, g_out = empty_tensor, empty_tensor

    return inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
        [x, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out, h_out, g_out]
    )


def sum(
    inputs: List[tf.Tensor],
    axis: int = -1,
    dc_decomp: bool = False,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    perturbation_domain: Optional[PerturbationDomain] = None,
    **kwargs: Any,
) -> List[tf.Tensor]:
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    mode = ForwardMode(mode)
    ibp = get_ibp(mode)
    affine = get_affine(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)

    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(
        inputs, compute_ibp_from_affine=False
    )
    dtype = x.dtype
    empty_tensor = inputs_outputs_spec.get_empty_tensor(dtype=dtype)

    if ibp:
        u_c_out = K.sum(u_c, axis=axis)
        l_c_out = K.sum(l_c, axis=axis)
    else:
        u_c_out, l_c_out = empty_tensor, empty_tensor

    if affine:
        if axis == -1:
            axis_w = -1
        else:
            axis_w = axis + 1
        w_u_out = K.sum(w_u, axis=axis_w)
        w_l_out = K.sum(w_l, axis=axis_w)
        b_u_out = K.sum(b_u, axis=axis)
        b_l_out = K.sum(b_l, axis=axis)
    else:
        w_u_out, b_u_out, w_l_out, b_l_out = empty_tensor, empty_tensor, empty_tensor, empty_tensor

    if dc_decomp:
        raise NotImplementedError("Not yet implemented for dc_decomp=True")
    else:
        h_out, g_out = empty_tensor, empty_tensor

    return inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
        [x, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out, h_out, g_out]
    )


def frac_pos(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    **kwargs: Any,
) -> List[tf.Tensor]:
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()

    mode = ForwardMode(mode)
    affine = get_affine(mode)
    ibp = get_ibp(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)

    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(inputs)
    dtype = x.dtype
    empty_tensor = inputs_outputs_spec.get_empty_tensor(dtype=dtype)

    u_c_out = 1.0 / l_c
    l_c_out = 1.0 / u_c

    if affine:
        w_u_0 = (u_c_out - l_c_out) / K.maximum(u_c - l_c, K.epsilon())
        b_u_0 = l_c_out - w_u_0 * l_c

        y = (u_c + l_c) / 2.0
        b_l_0 = 2.0 / y
        w_l_0 = -1 / y**2

        w_u_out = w_u_0[:, None] * w_l
        b_u_out = b_u_0 * b_l + b_u_0
        w_l_out = w_l_0[:, None] * w_u
        b_l_out = b_l_0 * b_u + b_l_0
    else:
        w_u_out, b_u_out, w_l_out, b_l_out = empty_tensor, empty_tensor, empty_tensor, empty_tensor

    if dc_decomp:
        raise NotImplementedError("Not yet implemented for dc_decomp=True")
    else:
        h_out, g_out = empty_tensor, empty_tensor

    return inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
        [x, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out, h_out, g_out]
    )


# convex hull of the maximum between two functions
def max_(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    axis: int = -1,
    finetune: bool = False,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """LiRPA implementation of max(x, axis)

    Args:
        inputs: list of tensors
        dc_decomp: boolean that indicates
        perturbation_domain: the type of perturbation domain
        axis: axis to perform the maximum
    whether we return a difference of convex decomposition of our layer

    Returns:
        max operation  along an axis
    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    mode = ForwardMode(mode)
    affine = get_affine(mode)
    ibp = get_ibp(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)
    inputs_outputs_spec_no_dc = InputsOutputsSpec(dc_decomp=False, mode=mode, perturbation_domain=perturbation_domain)

    input_shape = K.int_shape(inputs[-1])
    max_dim = input_shape[axis]
    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(
        inputs, tight=False, compute_ibp_from_affine=False
    )
    dtype = x.dtype
    empty_tensor = inputs_outputs_spec.get_empty_tensor(dtype=dtype)
    empty_tensors_list = [empty_tensor] * max_dim

    if mode == ForwardMode.IBP and not dc_decomp:
        u_c_out = K.max(u_c, axis=axis)
        l_c_out = K.max(l_c, axis=axis)

        return [u_c_out, l_c_out]

    # do some transpose so that the last axis is also at the end

    if ibp:
        u_c_list = tf.split(u_c, max_dim, axis)
        l_c_list = tf.split(l_c, max_dim, axis)
        u_c_tmp = u_c_list[0] + 0 * (u_c_list[0])
        l_c_tmp = l_c_list[0] + 0 * (l_c_list[0])
    else:
        u_c_tmp, l_c_tmp = empty_tensor, empty_tensor
        u_c_list, l_c_list = empty_tensors_list, empty_tensors_list

    if affine:
        b_u_list = tf.split(b_u, max_dim, axis)
        b_l_list = tf.split(b_l, max_dim, axis)
        b_u_tmp = b_u_list[0] + 0 * (b_u_list[0])
        b_l_tmp = b_l_list[0] + 0 * (b_l_list[0])

        if axis == -1:
            w_u_list = tf.split(w_u, max_dim, axis)
            w_l_list = tf.split(w_l, max_dim, axis)
        else:
            w_u_list = tf.split(w_u, max_dim, axis + 1)
            w_l_list = tf.split(w_l, max_dim, axis + 1)
        w_u_tmp = w_u_list[0] + 0 * (w_u_list[0])
        w_l_tmp = w_l_list[0] + 0 * (w_l_list[0])

        if finetune:
            params = kwargs["finetune_params"]
            params_split = [e[0] for e in tf.split(params[None], max_dim, axis)]
        else:
            params_split = [empty_tensor] * max_dim

    else:
        w_u_tmp, b_u_tmp, w_l_tmp, b_l_tmp = empty_tensor, empty_tensor, empty_tensor, empty_tensor
        w_u_list, b_u_list, w_l_list, b_l_list = (
            empty_tensors_list,
            empty_tensors_list,
            empty_tensors_list,
            empty_tensors_list,
        )
        params_split = [False] * max_dim

    h_tmp, g_tmp = None, None
    output_tmp = inputs_outputs_spec_no_dc.extract_outputsformode_from_fulloutputs(
        [
            x,
            u_c_tmp,
            w_u_tmp,
            b_u_tmp,
            l_c_tmp,
            w_l_tmp,
            b_l_tmp,
            h_tmp,
            g_tmp,
        ]
    )
    for i in range(1, max_dim):
        output_i = inputs_outputs_spec_no_dc.extract_outputsformode_from_fulloutputs(
            [x, u_c_list[i], w_u_list[i], b_u_list[i], l_c_list[i], w_l_list[i], b_l_list[i], None, None]
        )
        output_tmp = maximum(
            output_tmp,
            output_i,
            dc_decomp=False,
            mode=mode,
            finetune=finetune,
            finetune_params=params_split[i],
            perturbation_domain=perturbation_domain,
        )

    # reduce the dimension
    tight = not (mode == ForwardMode.AFFINE)  # no need to compute u_c, l_c for pure affine mode
    (
        _,
        u_c_out,
        w_u_out,
        b_u_out,
        l_c_out,
        w_l_out,
        b_l_out,
        _,
        _,
    ) = inputs_outputs_spec_no_dc.get_fullinputs_from_inputsformode(output_tmp, tight=tight)

    if mode in [ForwardMode.IBP, ForwardMode.HYBRID]:
        u_c_out = K.squeeze(u_c_out, axis)
        l_c_out = K.squeeze(l_c_out, axis)
    if mode in [ForwardMode.HYBRID, ForwardMode.AFFINE]:
        b_u_out = K.squeeze(b_u_out, axis)
        b_l_out = K.squeeze(b_l_out, axis)
        if axis == -1:
            w_u_out = K.squeeze(w_u_out, axis)
            w_l_out = K.squeeze(w_l_out, axis)
        else:
            w_u_out = K.squeeze(w_u_out, axis + 1)
            w_l_out = K.squeeze(w_l_out, axis + 1)

    if dc_decomp:
        g_out = K.sum(g, axis=axis)
        h_out = K.max(h + g, axis=axis) - g_out
    else:
        g_out, h_out = empty_tensor, empty_tensor

    return inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
        [x, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out, h_out, g_out]
    )


def softmax_to_linear(model: keras.Model) -> Tuple[keras.Model, bool]:
    """linearize the softmax layer for verification

    Args:
        model: Keras Model

    Returns:
        model without the softmax
    """
    layer = model.layers[-1]
    # check that layer is not an instance of the object Softmax
    if isinstance(layer, keras.layers.Softmax):
        model_normalize = keras.models.Model(model.get_input_at(0), keras.layers.Activation("linear")(layer.input))

        return model_normalize, True

    if hasattr(layer, "activation"):
        if not layer.get_config()["activation"] == "softmax":
            return model, False
        layer.get_config()["activation"] = "linear"
        layer.activation = keras.activations.get("linear")
        return model, True

    return model, False


def linear_to_softmax(model: keras.Model) -> Tuple[keras.Model, bool]:
    model.layers[-1].activation = keras.activations.get("softmax")
    return model


def multiply(
    inputs_0: List[tf.Tensor],
    inputs_1: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
) -> List[tf.Tensor]:
    """LiRPA implementation of (element-wise) multiply(x,y)=-x*y.

    Args:
        inputs_0: list of tensors
        inputs_1: list of tensors
        dc_decomp: boolean that indicates
        perturbation_domain: the type of perturbation domain
        mode: type of Forward propagation (ibp, affine, or hybrid)
    whether we return a difference of convex decomposition of our layer
    whether we propagate upper and lower bounds on the values of the gradient

    Returns:
        maximum(inputs_0, inputs_1)
    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()

    mode = ForwardMode(mode)
    affine = get_affine(mode)
    ibp = get_ibp(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)

    x_0, u_c_0, w_u_0, b_u_0, l_c_0, w_l_0, b_l_0, h_0, g_0 = inputs_outputs_spec.get_fullinputs_from_inputsformode(
        inputs_0
    )
    x_1, u_c_1, w_u_1, b_u_1, l_c_1, w_l_1, b_l_1, h_1, g_1 = inputs_outputs_spec.get_fullinputs_from_inputsformode(
        inputs_1
    )
    dtype = x_0.dtype
    empty_tensor = inputs_outputs_spec.get_empty_tensor(dtype=dtype)

    # using McCormick's inequalities to derive bounds
    # xy<= x_u*y + x*y_L - xU*y_L
    # xy<= x*y_u + x_L*y - x_L*y_U

    # xy >=x_L*y + x*y_L -x_L*y_L
    # xy >= x_U*y + x*y_U - x_U*y_U

    if ibp:
        u_c_0_out = K.maximum(u_c_0 * u_c_1, u_c_0 * l_c_1) + K.maximum(u_c_0 * l_c_1, l_c_0 * l_c_1) - u_c_0 * l_c_1
        u_c_1_out = K.maximum(u_c_1 * u_c_0, u_c_1 * l_c_0) + K.maximum(u_c_1 * l_c_0, l_c_1 * l_c_0) - u_c_1 * l_c_0
        u_c_out = K.minimum(u_c_0_out, u_c_1_out)
        l_c_out = K.minimum(l_c_0 * l_c_1, l_c_0 * u_c_1) + K.minimum(l_c_0 * l_c_1, u_c_0 * l_c_1) - l_c_0 * l_c_1
    else:
        u_c_out, l_c_out = empty_tensor, empty_tensor

    if affine:
        # xy <= x_u * y + x * y_L - xU * y_L
        cx_u_pos = K.maximum(u_c_0, 0.0)
        cx_u_neg = K.minimum(u_c_0, 0.0)

        cy_l_pos = K.maximum(l_c_1, 0.0)
        cy_l_neg = K.minimum(l_c_1, 0.0)
        w_u_out = (
            cx_u_pos[:, None] * w_u_1
            + cx_u_neg[:, None] * w_l_1
            + cy_l_pos[:, None] * w_u_0
            + cy_l_neg[:, None] * w_l_0
        )
        b_u_out = cx_u_pos * b_u_1 + cx_u_neg * b_l_1 + cy_l_pos * b_u_0 + cy_l_neg * b_l_0 - u_c_0 * l_c_1

        # xy >= x_U*y + x*y_U - x_U*y_U
        cx_l_pos = K.maximum(l_c_0, 0.0)
        cx_l_neg = K.minimum(l_c_0, 0.0)

        w_l_out = (
            cx_l_pos[:, None] * w_l_1
            + cx_l_neg[:, None] * w_u_1
            + cy_l_pos[:, None] * w_l_0
            + cy_l_neg[:, None] * w_u_0
        )
        b_l_out = cx_l_pos * b_l_1 + cx_l_neg * b_u_1 + cy_l_pos * b_l_0 + cy_l_neg * b_u_0 - l_c_0 * l_c_1
    else:
        w_u_out, b_u_out, w_l_out, b_l_out = empty_tensor, empty_tensor, empty_tensor, empty_tensor

    if dc_decomp:
        raise NotImplementedError("Not yet implemented for dc_decomp=True")
    else:
        h_out, g_out = empty_tensor, empty_tensor

    return inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
        [x_0, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out, h_out, g_out]
    )


def permute_dimensions(
    inputs: List[tf.Tensor],
    axis: int,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    axis_perm: int = 1,
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
) -> List[tf.Tensor]:
    """LiRPA implementation of (element-wise) permute(x,axis)

    Args:
        inputs: list of input tensors
        axis: axis on which we apply the permutation
        mode: type of Forward propagation (ibp, affine, or hybrid)
        axis_perm: see DecomonPermute operator

    Returns:

    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()

    mode = ForwardMode(mode)
    affine = get_affine(mode)
    ibp = get_ibp(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)

    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(
        inputs, tight=False, compute_ibp_from_affine=False
    )
    dtype = x.dtype
    empty_tensor = inputs_outputs_spec.get_empty_tensor(dtype=dtype)

    input_shape = inputs_outputs_spec.get_input_shape(inputs)
    if len(input_shape) <= 2:
        # not enough dim to permute
        return inputs
    index = np.arange(len(input_shape))
    index = np.insert(np.delete(index, axis), axis_perm, axis)
    index_w = np.arange(len(input_shape) + 1)
    index_w = np.insert(np.delete(index_w, axis), axis_perm + 1, axis)

    if ibp:
        u_c_out = K.permute_dimensions(u_c, index)
        l_c_out = K.permute_dimensions(l_c, index)
    else:
        u_c_out, l_c_out = empty_tensor, empty_tensor
    if affine:
        w_u_out = K.permute_dimensions(w_u, index_w)
        b_u_out = K.permute_dimensions(b_u, index)
        w_l_out = K.permute_dimensions(w_l, index_w)
        b_l_out = K.permute_dimensions(b_l, index)
    else:
        w_u_out, b_u_out, w_l_out, b_l_out = empty_tensor, empty_tensor, empty_tensor, empty_tensor

    if dc_decomp:
        raise NotImplementedError("Not yet implemented for dc_decomp=True")
    else:
        h_out, g_out = empty_tensor, empty_tensor

    return inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
        [x, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out, h_out, g_out]
    )


def broadcast(
    inputs: List[tf.Tensor],
    n: int,
    axis: int,
    mode: Union[str, ForwardMode],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
) -> List[tf.Tensor]:
    """LiRPA implementation of broadcasting

    Args:
        inputs
        n
        axis
        mode

    Returns:

    """
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    mode = ForwardMode(mode)
    affine = get_affine(mode)
    ibp = get_ibp(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)

    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(
        inputs, tight=False, compute_ibp_from_affine=False
    )
    dtype = x.dtype
    empty_tensor = inputs_outputs_spec.get_empty_tensor(dtype=dtype)

    if ibp:
        for _ in range(n):
            u_c = K.expand_dims(u_c, axis)
            l_c = K.expand_dims(l_c, axis)

    if axis != -1:
        axis_w = axis + 1
    else:
        axis_w = -1

    if affine:
        for _ in range(n):
            b_u = K.expand_dims(b_u, axis)
            b_l = K.expand_dims(b_l, axis)
            w_u = K.expand_dims(w_u, axis_w)
            w_l = K.expand_dims(w_l, axis_w)

    if dc_decomp:
        raise NotImplementedError("Not yet implemented for dc_decomp=True")

    return inputs_outputs_spec.extract_outputsformode_from_fulloutputs([x, u_c, w_u, b_u, l_c, w_l, b_l, h, g])


def split(
    inputs: List[tf.Tensor],
    axis: int = -1,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
) -> List[List[tf.Tensor]]:
    """LiRPA implementation of split

    Args:
        inputs
        axis
        mode

    Returns:

    """
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    mode = ForwardMode(mode)
    affine = get_affine(mode)
    ibp = get_ibp(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)

    input_shape = inputs_outputs_spec.get_input_shape(inputs)
    if axis == -1:
        n = input_shape[-1]
        axis = len(input_shape) - 1
    else:
        n = input_shape[axis]
    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(
        inputs, tight=False, compute_ibp_from_affine=False
    )
    dtype = x.dtype
    empty_tensor = inputs_outputs_spec.get_empty_tensor(dtype=dtype)
    empty_tensor_list = [empty_tensor] * n

    if ibp:
        u_c_list = tf.split(u_c, num_or_size_splits=n, axis=axis)
        l_c_list = tf.split(l_c, num_or_size_splits=n, axis=axis)
    else:
        u_c_list, l_c_list = empty_tensor_list, empty_tensor_list

    if affine:
        b_u_list = tf.split(b_u, num_or_size_splits=n, axis=axis)
        b_l_list = tf.split(b_l, num_or_size_splits=n, axis=axis)

        if axis != -1:
            axis += 1
        w_u_list = tf.split(w_u, num_or_size_splits=n, axis=axis)
        w_l_list = tf.split(w_l, num_or_size_splits=n, axis=axis)
    else:
        w_u_list, b_u_list, w_l_list, b_l_list = (
            empty_tensor_list,
            empty_tensor_list,
            empty_tensor_list,
            empty_tensor_list,
        )

    if dc_decomp:
        raise NotImplementedError("Not yet implemented for dc_decomp=True")
    else:
        h_list, g_list = empty_tensor_list, empty_tensor_list

    return [
        inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
            [x, u_c_list[i], w_u_list[i], b_u_list[i], l_c_list[i], w_l_list[i], b_l_list[i], h_list[i], g_list[i]]
        )
        for i in range(n)
    ]


def sort(
    inputs: List[tf.Tensor],
    axis: int = -1,
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
) -> List[tf.Tensor]:
    """LiRPA implementation of sort by selection

    Args:
        inputs
        axis
        dc_decomp
        perturbation_domain
        mode

    Returns:

    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()

    mode = ForwardMode(mode)
    affine = get_affine(mode)
    ibp = get_ibp(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)

    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(
        inputs, tight=False, compute_ibp_from_affine=False
    )
    input_shape = inputs_outputs_spec.get_input_shape(inputs)
    dtype = x.dtype
    empty_tensor = inputs_outputs_spec.get_empty_tensor(dtype=dtype)

    if axis == -1:
        n = input_shape[-1]
        axis = len(input_shape) - 1
    else:
        n = input_shape[axis]

    empty_tensor_list = [empty_tensor] * n

    # what about splitting elements
    op_split = lambda x: tf.split(x, n, axis=axis)
    if ibp:
        u_c_list = op_split(u_c)
        l_c_list = op_split(l_c)
    else:
        u_c_list, l_c_list = empty_tensor_list, empty_tensor_list

    if affine:
        w_u_list = tf.split(w_u, n, axis=axis + 1)
        b_u_list = op_split(b_u)
        w_l_list = tf.split(w_l, n, axis=axis + 1)
        b_l_list = op_split(b_l)
    else:
        w_u_list, b_u_list, w_l_list, b_l_list = (
            empty_tensor_list,
            empty_tensor_list,
            empty_tensor_list,
            empty_tensor_list,
        )

    if dc_decomp:
        raise NotImplementedError("Not yet implemented for dc_decomp=True")
    else:
        h_list, g_list = empty_tensor_list, empty_tensor_list

    # use selection sort
    for i in range(n - 1):
        for j in range(i + 1, n):
            input_i = inputs_outputs_spec.extract_inputsformode_from_fullinputs(
                [x, u_c_list[i], w_u_list[i], b_u_list[i], l_c_list[i], w_l_list[i], b_l_list[i], h_list[i], g_list[i]]
            )
            input_j = inputs_outputs_spec.extract_inputsformode_from_fullinputs(
                [x, u_c_list[j], w_u_list[j], b_u_list[j], l_c_list[j], w_l_list[j], b_l_list[j], h_list[j], g_list[j]]
            )

            # max and min of splitted inputs
            output_a = maximum(
                input_i, input_j, mode=mode, perturbation_domain=perturbation_domain, dc_decomp=dc_decomp
            )
            output_b = minimum(
                input_i, input_j, mode=mode, perturbation_domain=perturbation_domain, dc_decomp=dc_decomp
            )

            # update lists
            (
                x,
                u_c_list[j],
                w_u_list[j],
                b_u_list[j],
                l_c_list[j],
                w_l_list[j],
                b_l_list[j],
                h_list[j],
                g_list[j],
            ) = inputs_outputs_spec.get_fullinputs_from_inputsformode(output_a)
            (
                x,
                u_c_list[i],
                w_u_list[i],
                b_u_list[i],
                l_c_list[i],
                w_l_list[i],
                b_l_list[i],
                h_list[i],
                g_list[i],
            ) = inputs_outputs_spec.get_fullinputs_from_inputsformode(output_b)

    op_concatenate = lambda x: K.concatenate(x, axis)
    # update the inputs
    if ibp:
        u_c_out = op_concatenate(u_c_list)
        l_c_out = op_concatenate(l_c_list)
    else:
        u_c_out, l_c_out = empty_tensor, empty_tensor
    if affine:
        w_u_out = K.concatenate(w_u_list, axis + 1)
        w_l_out = K.concatenate(w_l_list, axis + 1)
        b_u_out = op_concatenate(b_u_list)
        b_l_out = op_concatenate(b_l_list)
    else:
        w_u_out, b_u_out, w_l_out, b_l_out = empty_tensor, empty_tensor, empty_tensor, empty_tensor

    if dc_decomp:
        raise NotImplementedError("Not yet implemented for dc_decomp=True")
    else:
        h_out, g_out = empty_tensor, empty_tensor

    return inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
        [x, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out, h_out, g_out]
    )


def pow(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
) -> List[tf.Tensor]:
    """LiRPA implementation of pow(x )=x**2

    Args:
        inputs
        dc_decomp
        perturbation_domain
        mode

    Returns:

    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    return multiply(inputs, inputs, dc_decomp=dc_decomp, perturbation_domain=perturbation_domain, mode=mode)


def abs(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
) -> List[tf.Tensor]:
    """LiRPA implementation of |x|

    Args:
        inputs
        dc_decomp
        perturbation_domain
        mode

    Returns:

    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    inputs_0 = relu_(inputs, dc_decomp=dc_decomp, perturbation_domain=perturbation_domain, mode=mode)
    inputs_1 = minus(
        relu_(
            minus(inputs, mode=mode, perturbation_domain=perturbation_domain),
            dc_decomp=dc_decomp,
            perturbation_domain=perturbation_domain,
            mode=mode,
        ),
        mode=mode,
        perturbation_domain=perturbation_domain,
    )

    return add(inputs_0, inputs_1, dc_decomp=dc_decomp, perturbation_domain=perturbation_domain, mode=mode)


def frac_pos_hull(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
) -> List[tf.Tensor]:
    """LiRPA implementation of 1/x for x>0

    Args:
        inputs
        dc_decomp
        perturbation_domain
        mode

    Returns:

    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()

    mode = ForwardMode(mode)
    affine = get_affine(mode)
    ibp = get_ibp(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)

    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(inputs)

    l_c = K.maximum(l_c, 1.0)
    z = (u_c + l_c) / 2.0
    w_l = -1 / K.pow(z)
    b_l = 2 / z
    w_u = (1.0 / u_c - 1.0 / l_c) / (u_c - l_c)
    b_u = 1.0 / u_c - w_u * u_c

    return [w_u, b_u, w_l, b_l]


# convex hull for min
def min_(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    axis: int = -1,
    finetune: bool = False,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """LiRPA implementation of min(x, axis=axis)

    Args:
        inputs
        dc_decomp
        perturbation_domain
        mode
        axis

    Returns:

    """
    # return - max - x
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    return minus(
        max_(
            minus(inputs, mode=mode, perturbation_domain=perturbation_domain),
            dc_decomp=dc_decomp,
            perturbation_domain=perturbation_domain,
            mode=mode,
            axis=axis,
            finetune=finetune,
            **kwargs,
        ),
        mode=mode,
        perturbation_domain=perturbation_domain,
    )


def expand_dims(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    axis: int = -1,
    perturbation_domain: Optional[PerturbationDomain] = None,
    **kwargs: Any,
) -> List[tf.Tensor]:
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    mode = ForwardMode(mode)
    affine = get_affine(mode)
    ibp = get_ibp(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)

    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(
        inputs, tight=False, compute_ibp_from_affine=False
    )
    dtype = x.dtype
    empty_tensor = inputs_outputs_spec.get_empty_tensor(dtype=dtype)

    if axis == -1:
        axis_w = axis
    else:
        axis_w = axis + 1

    op = lambda t: K.expand_dims(t, axis)

    if ibp:
        u_c_out = op(u_c)
        l_c_out = op(l_c)
    else:
        u_c_out, l_c_out = empty_tensor, empty_tensor

    if affine:
        b_u_out = op(b_u)
        b_l_out = op(b_l)
        w_u_out = K.expand_dims(w_u, axis_w)
        w_l_out = K.expand_dims(w_l, axis_w)
    else:
        w_u_out, b_u_out, w_l_out, b_l_out = empty_tensor, empty_tensor, empty_tensor, empty_tensor

    if dc_decomp:
        raise NotImplementedError("Not yet implemented for dc_decomp=True")
    else:
        h_out, g_out = empty_tensor, empty_tensor

    return inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
        [x, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out, h_out, g_out]
    )


def log(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """Exponential activation function.

    Args:
        inputs: list of input tensors
        dc_decomp: boolean that indicates
        perturbation_domain: the type of convex input domain
        mode: type of Forward propagation (ibp, affine, or hybrid)
        **kwargs: extra parameters
    whether we return a difference of convex decomposition of our layer

    Returns:
        the updated list of tensors
    """
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()

    mode = ForwardMode(mode)
    affine = get_affine(mode)
    ibp = get_ibp(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)

    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(inputs)
    dtype = x.dtype
    empty_tensor = inputs_outputs_spec.get_empty_tensor(dtype=dtype)

    if mode == ForwardMode.AFFINE:
        l_c = K.maximum(K.cast(K.epsilon(), dtype=l_c.dtype), l_c)

    u_c_out = K.log(u_c)
    l_c_out = K.log(l_c)

    if affine:
        y = (u_c + l_c) / 2.0

        w_l_0 = (u_c_out - l_c_out) / K.maximum(u_c - l_c, K.epsilon())
        b_l_0 = l_c_out - w_l_0 * l_c

        w_u_0 = 1 / y
        b_u_0 = K.log(y) - 1

        w_u_out = w_u_0[:, None] * w_u
        b_u_out = w_u_0 * b_u + b_u_0
        w_l_out = w_l_0[:, None] * w_l
        b_l_out = w_l_0 * b_l + b_l_0
    else:
        w_u_out, b_u_out, w_l_out, b_l_out = empty_tensor, empty_tensor, empty_tensor, empty_tensor

    if dc_decomp:
        raise NotImplementedError("Not yet implemented for dc_decomp=True")
    else:
        h_out, g_out = empty_tensor, empty_tensor

    return inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
        [x, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out, h_out, g_out]
    )


def exp(
    inputs: List[tf.Tensor],
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """Exponential activation function.

    Args:
        inputs: list of input tensors
        dc_decomp: boolean that indicates
        perturbation_domain: the type of convex input domain
        mode: type of Forward propagation (ibp, affine, or hybrid)
        slope:
        **kwargs: extra parameters
    whether we return a difference of convex decomposition of our layer

    Returns:
        the updated list of tensors
    """
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()

    mode = ForwardMode(mode)
    affine = get_affine(mode)
    ibp = get_ibp(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)

    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(inputs)
    dtype = x.dtype
    empty_tensor = inputs_outputs_spec.get_empty_tensor(dtype=dtype)

    u_c_out = K.exp(u_c)
    l_c_out = K.exp(l_c)

    if affine:
        y = (u_c + l_c) / 2.0  # do finetuneting

        w_u_0 = (u_c_out - l_c_out) / K.maximum(u_c - l_c, K.epsilon())
        b_u_0 = l_c_out - w_u_0 * l_c

        w_l_0 = K.exp(y)
        b_l_0 = w_l_0 * (1 - y)

        w_u_out = w_u_0[:, None] * w_u
        b_u_out = w_u_0 * b_u + b_u_0
        w_l_out = w_l_0[:, None] * w_l
        b_l_out = w_l_0 * b_l + b_l_0
    else:
        w_u_out, b_u_out, w_l_out, b_l_out = empty_tensor, empty_tensor, empty_tensor, empty_tensor

    if dc_decomp:
        raise NotImplementedError("Not yet implemented for dc_decomp=True")
    else:
        h_out, g_out = empty_tensor, empty_tensor

    return inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
        [x, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out, h_out, g_out]
    )
